import random


class IntervalTree:
    """
    Binary search tree for storing long integer intervals, and for performing queries on them.
    See https://en.wikipedia.org/wiki/Interval_tree, specifically the Augmented kind.
    The present implementation balances the tree by using randomisation.
    """
    root = None # pointer to the root node of the binary search tree
    stack = None

    def __iter__(self):
        self.current = self.root
        self.stack = Stack()
        return self

    def __next__(self):
        while self.current != None:
            self.stack.push(self.current)
            self.current = self.current.left
        if self.stack.isEmpty():
            raise StopIteration
        self.current = self.stack.pop()
        ret = self.current
        self.current = self.current.right
        return ret

    def __len__(self):
        return self.root.N

    def __contains__(self, ival):
        return self.get(ival) != None

    def get(self, ival):
        return self._get(self.root, ival)

    def _get(self, node, ival):
        if node == None: return None
        if ival < node.ival:
            return self._get(node.left, ival)
        elif ival > node.ival:
            return self._get(node.right, ival)
        else:
            return node

    def isect(self, ival, node = None):
        """ Look for intersecting interval in subtree rooted at specified node (root by default).
            Returns node of intersecting interval. """
        if self.root == None: return None
        if node == None: return self.isect(ival, self.root)
        while node != None:
            if isect(ival, node.ival): return node
            elif node.left == None: node = node.right
            elif node.left.max < ival.min: node = node.right
            else: node = node.left
        return None

    def isectall(self, ival):
        """ Look for all intersecting intervals in the subtree rooted at specified node (root by default).
            Returns nodes of intersecting intervals. """
        return _isectall(ival, self.root)

    def closest(self, query):
        """ Retrieve the interval Y stored in the tree that is closest to the given interval X.
            If the given interval overlaps with one or more stored intervals, one is returned:
            the interval Y with the greatest Jaccard index to X. If multiple intervals are equally close,
            only one is returned (the one before I think).
            :param query: the interval for which the closest is sought
            :return: the interval closest to the given query interval
        """
        ovlap = self.isectall(query)
        if len(ovlap) == 0: # overlapping intervals are not in the tree
            return _closest(query, self.root)
        else:
            best_iv = None
            best_ji = 0
            for node in ovlap:
                ji = jaccard(node.ival, query)
                if best_iv == None or ji > best_ji:
                    best_iv = node
                    best_ji = ji
            return best_iv

    def put(self, ival, value = None):
        nodex = self.get(ival)
        if nodex:
            nodex.values.add(value)
        else:
            self.root = self._randomizedInsert(self.root, ival, value)

    def _randomizedInsert(self, node, ival, value):
        if node == None: return IntervalNode(ival, value)
        if random.uniform(0,1) * node.N < 1.0: return self._rootInsert(node, ival, value)
        if ival < node.ival:
            node.left = self._randomizedInsert(node.left, ival, value)
        else:
            node.right = self._randomizedInsert(node.right, ival, value)
        _fix(node)
        return node

    def _rootInsert(self, node, ival, value):
        if node == None: return IntervalNode(ival, value)
        if ival < node.ival:
            node.left = self._rootInsert(node.left, ival, value)
            node = _rotR(node)
        else:
            node.right = self._rootInsert(node.right, ival, value)
            node = _rotL(node)
        return node



def _isectall(ival, node):
    """ Look for all intersecting intervals in the subtree rooted at specified node (root by default).
        Returns nodes of intersecting intervals. """
    if node == None: return []
    found = []
    if isect(ival, node.ival):
        found = [node]
    if node.left and node.left.max >= ival.min:
        found.extend(_isectall(ival, node.left))
    if len(found) > 0 or node.left == None or node.left.max < ival.min:
        found.extend(_isectall(ival, node.right))
    return found

def _closest(query, cand):
    """ Recursively find the interval with the minimum distance to that given.
        This internal function does not guarantee that distances are sensible when overlapping
        intervals exist; essentially it assumes that overlaps have been eliminated prior.
        :param query: interval
        :param cand: node from which search starts
        :return: closest interval """
    fav = None
    favdist = -1
    while cand != None:
        if query == cand.ival: return cand
        distx = query.dist(cand.ival)
        if fav == None or distx <= favdist:
            fav = cand
            favdist = distx
        if cand.left == None: cand = cand.right
        elif cand.right == None: cand = cand.left
        elif cand.ival.min > query.max: cand = cand.left # the smallest, indexed value (on left) is AFTER the query min
        else: # no way to choose without looking in the intervals below
            favleft = None
            distleft = query.dist(Interval(cand.left.min, cand.left.max))
            if distleft < favdist:
                favleft = _closest(query, cand.left)
                distleft = query.dist(favleft.ival) if favleft != None else MAX_VALUE
            distright = query.dist(Interval(cand.right.min, cand.right.max))
            if distright < favdist:
                favright = _closest(query, cand.right)
                distright = query.dist(favright.ival) if favright != None else MAX_VALUE
            if distleft < distright:
                return favleft if distleft < favdist else fav
            else:
                return favright if distright < favdist else fav
    return fav

class IntervalNode:
    """
    Defines the node of the interval search tree.
    Manages values associated with intervals.
    """
    ival = None     # the actual interval
    values = None   # values associated with the interval
    left = None     # subtree on the left (lesser)
    right = None    # subtree on the right (greater)
    N = 1           # number of nodes under (and including) this one
    min = 0         # min point of subtree rooted at this node
    max = 0         # max point of subtree rooted at this node

    def __init__(self, interval, value = None):
        self.ival = interval
        self.min = interval.min
        self.max = interval.max
        self.values = set()
        if value != None:
            self.values.add(value)

    def add(self, value):
        if value:
            self.values.add(value)

    def __str__(self):
        leftstr = 'o' if self.left else 'x'
        rightstr = 'o' if self.right else 'x'
        return leftstr + self.ival.__str__() + rightstr

    def __unicode__(self):
        leftstr = 'o' if self.left else 'x'
        rightstr = 'o' if self.right else 'x'
        return leftstr + self.ival.__unicode__() + rightstr

def size(node):
    if node == None: return 0
    else: return node.N

def _fix(node):
    if node == None: return
    node.N = 1 + size(node.left) + size(node.right)
    node.min = _min3(node.ival.min, _min(node.left), _min(node.right))
    node.max = _max3(node.ival.max, _max(node.left), _max(node.right))

MAX_VALUE = 9E30
MIN_VALUE = -9E30

def _min(node):
    return MAX_VALUE if node == None else node.min

def _max(node):
    return MIN_VALUE if node == None else node.max

def _min3(a, b, c):
    return min(a, min(b, c))

def _max3(a, b, c):
    return max(a, max(b, c))

def _rotR(node):
    y = node.left
    node.left = y.right
    y.right = node
    _fix(node)
    _fix(y)
    return y

def _rotL(node):
    y = node.right
    node.right = y.left
    y.left = node
    _fix(node)
    _fix(y)
    return y

class Stack:
    """ A stack to support an iterator over IntervalNodes in the IntervalTree. """
    def __init__(self):
        self.items = []

    def isEmpty(self):
        return self.items == []

    def push(self, item):
        self.items.append(item)

    def pop(self):
        return self.items.pop()

    def peek(self):
        return self.items[len(self.items) - 1]

    def size(self):
        return len(self.items)

class Interval:
    """
    Define a one-dimensional interval.
    """
    def __init__(self, min, max):
        if (min <= max):
            self.min = min
            self.max = max
        else:
            raise RuntimeError

    def isect(self, that):
        if (that.max < self.min): return False
        if (self.max < that.min): return False
        return True

    def isectStrict(self, that):
        if (that.max <= self.min): return False
        if (self.max <= that.min): return False
        return True

    def contains(self, x):
        return (min <= x) and (x <= max)

    def __eq__(self, other):
        if not isinstance(other, Interval): return False
        return True if (self.min == other.min and self.max == other.max) else False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, Interval): return False
        return True if (self.min < other.min or (self.min == other.min and self.max < other.max)) else False

    def __gt__(self, other):
        if not isinstance(other, Interval): return False
        return True if (self.min > other.min or (self.min == other.min and self.max > other.max)) else False

    def __str__(self):
        return '[' + str(self.min) + ', ' + str(self.max) +']'

    def __unicode__(self):
        return '[' + str(self.min) + ', ' + str(self.max) +']'

    def __sizeof__(self):
        return self.max - self.min

    def dist(self, that, signed = False, centre2centre = False):
        """ Calculate and return the closest distance (from one end of the interval of this to the end of the interval of that).
            If centre2centre is True, use the centre-to-centre distance instead.
            If signed is True, the distance is negative if this interval is after the that.
        """
        if not centre2centre:
            if not signed:
                if (self.min > that.max): return self.min - that.max # that interval is BEFORE this
                if (self.max < that.min): return that.min - self.max # that interval is AFTER this
            else: # distance is signed
                if (self.min > that.max): return that.max - self.min # that interval is BEFORE this
                if (self.max < that.min): return that.min - self.max # that interval is AFTER this
            return 0
        else:
            thiscentre = (self.max - self.min) / 2 + self.min
            thatcentre = (that.max - that.min) / 2 + that.min
            return thatcentre - thiscentre if signed else abs(thatcentre - thiscentre)

def dist(first, second, signed = False, centre2centre = False):
    """ Calculate and return the closest distance (from one end of the interval to the other).
        If centre2centre is True, use the centre-to-centre distance instead.
        If signed is True, the distance is negative if the first is after the second.
    """
    if isinstance(first, Interval) and isinstance(second, Interval):
        return first.dist(second, signed, centre2centre)
    return RuntimeError

def union(first, second):
    if (first.isect(second)):
        min = first.min if (first.min < second.min) else second.min
        max = second.max if (first.max < second.max) else first.max
        return Interval(min, max)
    else:
        raise RuntimeError

def isect(first, second):
    if (first.isect(second)):
        min = first.min if (first.min > second.min) else second.min
        max = second.max if (first.max > second.max) else first.max
        return Interval(min, max)
    else:
        return None

def jaccard(first, second):
        if (isect(first, second)):
            isect_min = first.min if (first.min > second.min) else second.min
            isect_max = second.max if (first.max > second.max) else first.max
            union_min = first.min if (first.min < second.min) else second.min
            union_max = second.max if (first.max < second.max) else first.max
            denom = union_max - union_min
            if (denom > 0):
                return (isect_max - isect_min) / denom
            return 0
        else:
            return 0

if __name__ == '__main__':
    a = Interval(13, 20)
    b = Interval(25, 30)
    c = Interval(27, 33)
    d = Interval(40, 50)
    e = Interval(21, 22)
    f = Interval(36, 38)
    g = Interval(16, 19)
    h = Interval(28, 31)
    i = Interval(55, 66)
    j = Interval(-3,  0)
    k = Interval(24, 24)
    l = Interval(52, 55)

    print('dist(b,a,signed=False,centre2centre=False)=', dist(b, a, signed = False, centre2centre=False))
    print('dist(b,a,signed=True,centre2centre=False)=', dist(b, a, signed = True, centre2centre=False))
    print('dist(b,a,signed=False,centre2centre=True)=', dist(b, a, signed = False, centre2centre=True))
    print('dist(b,a,signed=True,centre2centre=True)=', dist(b, a, signed = True, centre2centre=True))
    t = IntervalTree()
    t.put(a, 'A')
    t.put(b, 'B')
    t.put(c, 'C')
    t.put(d, 'D')
    t.put(e, 'E')
    t.put(b, 123)
    t.put(b, 'blah')
    t.get(d).add('x999')
    t.put(i)
    t.put(j)
    t.put(g)
    t.put(k)

    print(c in t)
    print(e in t)
    print(t.get(a).values)
    print(t.get(d).values)
    print(t.get(b).values)

    print(t.isect(f))
    print(t.isect(g))

    tryme = f

    all = t.isectall(tryme)
    print("Intersect with " + str(tryme) + ": ")
    for n in all:
        print('\t' + str(n))

    print("Closest to " + str(tryme) + ": ")
    print(t.closest(tryme))

    print('Iterate through tree: ')
    for n in t:
        print('\t' + str(n))


#!/usr/bin/python3

import argparse
import sys
from chipr import rankprod, bed



class RankProdAnalysis(object):

    def __init__(self):
        parser = argparse.ArgumentParser(prog='chipr',
                                         description="Combine multiple ChIP-seq files and return a union of all peak "
                                                     "locations and a set confident, reproducible peaks as determined by "
                                                     "rank product analysis")
        parser.add_argument("-i", "--input",
                            help="ChIP-seq input files. These files must be in either narrowPeak, broadPeak, "
                                 "or regionPeak format. Multiple inputs are separeted by a single space",
                            dest="input",
                            type=argparse.FileType('r'),
                            nargs='+',
                            required=True)
        parser.add_argument("-o", "--output",
                            help="ChIP-seq output filename prefix",
                            dest="output",
                            type=str,
                            default="rankprod",
                            required=False)
        parser.set_defaults(bigbed=False)
        parser.add_argument("-m", "--minentries",
                            help="The minimum peaks between replicates required to form an "
                                 "intersection of the peaks \n"
                                 "Default: 1",
                            dest="minentries",
                            default=1,
                            type=int,
                            required=False)
        parser.add_argument("--rankmethod",
                            help="The ranking method used to rank peaks within replicates. "
                                 "Options: 'signalvalue', 'pvalue', 'qvalue'. \n"
                                 "Default: pvalue",
                            dest="rankmethod",
                            default='pvalue',
                            type=str,
                            required=False)
        parser.set_defaults(broadpeaks=False)
        parser.add_argument("--fragment",
                            help="Specifies whether the input peaks will be subject to high levels of fragmentation",
                            dest="fragment",
                            action="store_true",
                            required=False)
        parser.add_argument("--duphandling",
                            help="Specifies how to handle entries that are ranked equally within a replicate "
                                 "Can either take the 'average' ranks or a 'random' rearrangement of the ordinal ranks \n"
                                 "Options: 'average', 'random' \n"
                                 "Default: 'average'",
                            dest="duphandling",
                            default='average',
                            type=str,
                            required=False)
        parser.add_argument("--seed",
                            help="Specify a seed to be used in conjunction with the 'random' option for -duphandling "
                                 "Must be between 0 and 1 \n"
                                 "Default: 0.5",
                            dest="random_seed",
                            default=0.5,
                            type=float,
                            required=False)
        parser.add_argument("-a","--alpha",
                            help="Alpha specifies the user cut-off value for set of reproducible peaks "
                                 "The analysis will still produce results including peaks within the threshold calculated"
                                 "using the binomial method \n"
                                 "Default: 0.05",
                            dest="alpha",
                            default=0.05,
                            type=float,
                            required=False)
        parser.add_argument("-s", "--size",
                            help="Sets the default minimum peak size when peaks are reconnected after fragmentation. \n"
                                 "Usually the minimum peak size is determined by the size of surrounding peaks, \n "
                                 "but in the case that there are no surrounding peaks this value will be used \n"
                                 "Default: 20",
                            dest="size",
                            default=20,
                            type=int,
                            required=False)
        args = parser.parse_args(sys.argv[1:])
        self.run_1(args)

    def run_1(self, args):
        for i in args.input:
            print(str(i.name))

        print('Processing Input...')

        bedfs = [bed.BedFile(str(i.name), 'Peaks') for i in args.input]

        rankprod.performrankprod(bedfs,
                                 args.minentries,
                                 args.rankmethod,
                                 'all',
                                 args.duphandling,
                                 args.random_seed,
                                 args.alpha,
                                 args.output,
                                 args.size,
                                 False,
                                 args.fragment)



def main():
    RankProdAnalysis()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
import scipy as sp
from scipy import interpolate


def _ecdf(x):
    '''no frills empirical cdf used in fdrcorrection
    '''
    nobs = len(x)
    return np.arange(1,nobs+1)/float(nobs)

'''
Multiple Testing and P-Value Correction


Author: Josef Perktold
License: BSD-3

'''


def multipletests(pvals, alpha=0.05, method='b', is_sorted=False,
                  returnsorted=False):
    """
    Test results and p-value correction for multiple tests

    Parameters
    ----------
    pvals : array_like, 1-d
        uncorrected p-values.   Must be 1-dimensional.
    alpha : float
        FWER, family-wise error rate, e.g. 0.1
    method : string
        Method used for testing and adjustment of pvalues. Can be either the
        full name or initial letters. Available methods are:

        - `bonferroni` : one-step correction
        - `holm` : step-down method using Bonferroni adjustments

    is_sorted : bool
        If False (default), the p_values will be sorted, but the corrected
        pvalues are in the original order. If True, then it assumed that the
        pvalues are already sorted in ascending order.
    returnsorted : bool
         not tested, return sorted p-values instead of original sequence

    Returns
    -------
    reject : array, boolean
        true for hypothesis that can be rejected for given alpha
    pvals_corrected : array
        p-values corrected for multiple tests
    alphacSidak: float
        corrected alpha for Sidak method
    alphacBonf: float
        corrected alpha for Bonferroni method

    Notes
    -----
    **Timing**:

    Most of the time with large arrays is spent in `argsort`. When
    we want to calculate the p-value for several methods, then it is more
    efficient to presort the pvalues, and put the results back into the
    original order outside of the function.

    """
    import gc
    pvals = np.asarray(pvals)
    alphaf = alpha  # Notation ?

    if not is_sorted:
        sortind = np.argsort(pvals)
        pvals = np.take(pvals, sortind)

    ntests = len(pvals)
    alphacSidak = 1 - np.power((1. - alphaf), 1./ntests)
    alphacBonf = alphaf / float(ntests)
    if method.lower() in ['b', 'bonf', 'bonferroni']:
        reject = pvals <= alphacBonf
        pvals_corrected = pvals * float(ntests)

    elif method.lower() in ['h', 'holm']:
        notreject = pvals > alphaf / np.arange(ntests, 0, -1)
        nr_index = np.nonzero(notreject)[0]
        if nr_index.size == 0:
            # nonreject is empty, all rejected
            notrejectmin = len(pvals)
        else:
            notrejectmin = np.min(nr_index)
        notreject[notrejectmin:] = True
        reject = ~notreject
        pvals_corrected_raw = pvals * np.arange(ntests, 0, -1)
        pvals_corrected = np.maximum.accumulate(pvals_corrected_raw)
        del pvals_corrected_raw
        gc.collect()

    else:
        raise ValueError('method not recognized')

    if not pvals_corrected is None: #not necessary anymore
        pvals_corrected[pvals_corrected>1] = 1
    if is_sorted or returnsorted:
        return reject, pvals_corrected, alphacSidak, alphacBonf
    else:
        pvals_corrected_ = np.empty_like(pvals_corrected)
        pvals_corrected_[sortind] = pvals_corrected
        del pvals_corrected
        reject_ = np.empty_like(reject)
        reject_[sortind] = reject
        return reject_, pvals_corrected_, alphacSidak, alphacBonf


def fdrcorrection(pvals, alpha=0.05, method='indep', is_sorted=False):
    '''pvalue correction for false discovery rate

    This covers Benjamini/Hochberg for independent or positively correlated and
    Benjamini/Yekutieli for general or negatively correlated tests. Both are
    available in the function multipletests, as method=`fdr_bh`, resp. `fdr_by`.

    Parameters
    ----------
    pvals : array_like
        set of p-values of the individual tests.
    alpha : float
        error rate
    method : {'indep', 'negcorr')

    Returns
    -------
    rejected : array, bool
        True if a hypothesis is rejected, False if not
    pvalue-corrected : array
        pvalues adjusted for multiple hypothesis testing to limit FDR

    Notes
    -----

    If there is prior information on the fraction of true hypothesis, then alpha
    should be set to alpha * m/m_0 where m is the number of tests,
    given by the p-values, and m_0 is an estimate of the true hypothesis.
    (see Benjamini, Krieger and Yekuteli)

    The two-step method of Benjamini, Krieger and Yekutiel that estimates the number
    of false hypotheses will be available (soon).

    Method names can be abbreviated to first letter, 'i' or 'p' for fdr_bh and 'n' for
    fdr_by.



    '''
    pvals = np.asarray(pvals)

    if not is_sorted:
        pvals_sortind = np.argsort(pvals)
        pvals_sorted = np.take(pvals, pvals_sortind)
    else:
        pvals_sorted = pvals  # alias

    if method in ['i', 'indep', 'p', 'poscorr']:
        ecdffactor = _ecdf(pvals_sorted)
    elif method in ['n', 'negcorr']:
        cm = np.sum(1./np.arange(1, len(pvals_sorted)+1))   #corrected this
        ecdffactor = _ecdf(pvals_sorted) / cm
##    elif method in ['n', 'negcorr']:
##        cm = np.sum(np.arange(len(pvals)))
##        ecdffactor = ecdf(pvals_sorted)/cm
    else:
        raise ValueError('only indep and negcorr implemented')
    reject = pvals_sorted <= ecdffactor*alpha
    if reject.any():
        rejectmax = max(np.nonzero(reject)[0])
        reject[:rejectmax] = True

    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
    del pvals_corrected_raw
    pvals_corrected[pvals_corrected>1] = 1
    if not is_sorted:
        pvals_corrected_ = np.empty_like(pvals_corrected)
        pvals_corrected_[pvals_sortind] = pvals_corrected
        del pvals_corrected
        reject_ = np.empty_like(reject)
        reject_[pvals_sortind] = reject
        return reject_, pvals_corrected_
    else:
        return reject, pvals_corrected


def estimate(pv, m=None, verbose=False, lowmem=False, pi0=None):
    """
    Estimates q-values from p-values
    Args
    =====
    m: number of tests. If not specified m = pv.size
    verbose: print verbose messages? (default False)
    lowmem: use memory-efficient in-place algorithm
    pi0: if None, it's estimated as suggested in Storey and Tibshirani, 2003.
         For most GWAS this is not necessary, since pi0 is extremely likely to be
         1
    """
    assert(pv.min() >= 0 and pv.max() <= 1), "p-values should be between 0 and 1"

    original_shape = pv.shape
    pv = pv.ravel()  # flattens the array in place, more efficient than flatten()

    if m is None:
        m = float(len(pv))
    else:
        # the user has supplied an m
        m *= 1.0

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100 and pi0 is None:
        pi0 = 1.0
    elif pi0 is not None:
        pi0 = pi0
    else:
        # evaluate pi0 for different lambdas
        pi0 = []
        lam = sp.arange(0, 0.90, 0.01)
        counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])
        for l in range(len(lam)):
            pi0.append(counts[l]/(m*(1-lam[l])))

        pi0 = sp.array(pi0)

        # fit natural cubic spline
        tck = interpolate.splrep(lam, pi0, k=3)
        pi0 = interpolate.splev(lam[-1], tck)
        if verbose:
            print("qvalues pi0=%.3f, estimated proportion of null features " % pi0)

        if pi0 > 1:
            if verbose:
                print("got pi0 > 1 (%.3f) while estimating qvalues, setting it to 1" % pi0)
            pi0 = 1.0

    assert(pi0 >= 0 and pi0 <= 1), "pi0 is not between 0 and 1: %f" % pi0

    if lowmem:
        # low memory version, only uses 1 pv and 1 qv matrices
        qv = sp.zeros((len(pv),))
        last_pv = pv.argmax()
        qv[last_pv] = (pi0*pv[last_pv]*m)/float(m)
        pv[last_pv] = -sp.inf
        prev_qv = last_pv
        for i in xrange(int(len(pv))-2, -1, -1):
            cur_max = pv.argmax()
            qv_i = (pi0*m*pv[cur_max]/float(i+1))
            pv[cur_max] = -sp.inf
            qv_i1 = prev_qv
            qv[cur_max] = min(qv_i, qv_i1)
            prev_qv = qv[cur_max]

    else:
        p_ordered = sp.argsort(pv)
        pv = pv[p_ordered]
        qv = pi0 * m/len(pv) * pv
        qv[-1] = min(qv[-1], 1.0)

        for i in range(len(pv)-2, -1, -1):
            qv[i] = min(pi0*m*pv[i]/(i+1.0), qv[i+1])

        # reorder qvalues
        qv_temp = qv.copy()
        qv = sp.zeros_like(qv)
        qv[p_ordered] = qv_temp

    # reshape qvalues
    qv = qv.reshape(original_shape)

    return qv



from chipr import bed
import scipy
import numpy as np
from chipr import rankprod
from chipr import multipletesting

"""
Dead code soon to be deleted
"""
def verifyChIAPeaks(chiaData, chipData, filename, alpha):
    listinters = []
    PETs = []
    probs = []
    inters = {}
    PETgroups = {}
    prs = []
    fdr = []
    RP = []
    if getattr(chiaData, 'format') == 'BED12':
        chiaData = bed.BED12toBEDPE(chiaData)
    for link in chiaData:
        if link.PETs <= 2:
            continue
        o1 = chipData.getOverlap(link.partner1)
        o2 = chipData.getOverlap(link.partner2)
        try:
            PETgroups[link.PETs].append(link)
        except KeyError:
            PETgroups[link.PETs] = [link]

        try:
            if len(o1) > 0 and len(o2) > 0:
                for peaks1, peaks2 in zip(o1, o2):
                    # Create entries in dictionary with peaks as support
                    try:

                        inters[link].append([peaks1, peaks2])
                    except KeyError:

                        inters[link] = [peaks1, peaks2]
        except TypeError:
            continue
    dists = []
    for k, v in inters.items():
        pvals = []
        rs = []
        # if k.PETs <= 1:
        #     continue
        # kvals = []
        for peak in v:
            try:
                pvals.append(peak.pValue)
                rs.append(peak.signalValue)

                # combined = scipy.stats.combine_pvalues([peak1.pValue, peak2.pValue], 'fisher')
                # pvals.append(combined[1])
            except AttributeError:
                if type(peak) == 'list':
                    for p in peak:
                        try:
                            pvals.append(p.pValue)
                            rs.append(peak.signalValue)
                        except AttributeError:
                            continue
                else:
                    continue
        if len(pvals) != 0:
            combined = scipy.stats.combine_pvalues(pvals, 'stouffer', rs)
            probs.append(combined[1])
            # rs = iterflatten(rs)
            # RP.append(rs)
            listinters.append(k)
            PETs.append(k.PETs)
            dists.append(k.getDistance())

    # probcor = multipletesting.fdrcorrection(probs, alpha)
    PETs, listinters, probcor, dists = (list(x) for x in zip(*sorted(zip(PETs, listinters, probs, dists),
                                                              key=lambda pair: pair[0], reverse=True)))

    bins = np.arange(min(dists), max(dists), max(dists)/50, dtype=int)
    brobs = []
    ints = []
    pets2 = []
    ainters = []
    tags = 0

    n = len(RP)
        # k = sum(pets)
    t = sum(PETs)
        # brobs = []
    tags = 0

    for i, p in enumerate(zip(probcor, PETs)):
        tags += p[1]
        b = scipy.stats.binom.cdf(tags, t, 1-p[0])
        print(p[0], p[1])
        brobs.append(b)
        pets2.append(p[1])
        ints.append(listinters[i])

    corrected = multipletesting.fdrcorrection(brobs, alpha)
    thresh = rankprod.thresholdCalc(brobs)
    try:
        binomAlpha = round(min(thresh[2]), 3)
        if binomAlpha==0:
            binomAlpha=0.05
    except ValueError:
        binomAlpha = 0.05
    # brobs.append(rpb)
    print(binomAlpha)
    for i, p in zip(ints, corrected[1]):
        i.addOption(pValue=p)
        prs.append(p)
        ainters.append(i)
        # print(p, binomAlpha)
        if p <= 0.05:
            if p == 0:
                i.addOption(pValue=0.00000000000001)
            fdr.append(i)

    fdr = bed.BedFile(fdr, "BEDPE")
    allinters = bed.BedFile(ainters, "BEDPE")
    print(len(fdr), 'Interactions pass FDR threshold')
    bed.writeBedFile(allinters, filename + '.bedpe', format='BEDPE')
    bed.writeBedFile(fdr, filename + '.bedpe', format="BEDPE")
    intersBED12 = bed.BEDPEtoBED12(allinters)
    bed.writeBedFile(intersBED12, filename + '.bed', format='BED12')

    return ainters, fdr, allinters, PETs, dists, brobs, corrected[1]


'''
joinChIA:
Helper function that puts all the interactions from multiple bedPE files in to one file
'''


def joinChIA(chia):
    interactions = []
    for rep in chia:
        for intr in rep:
            interactions.append(intr)

    pe = bed.BedFile(interactions, 'bedpe')
    return pe


def ChIAreproducibility(chia, chip, minentries=None, rank='signalValue', threshold='all',
                        alpha=0.05, filename='NA'):
    if minentries is None:
        minentries = len(chip)
    if len(chia) == 1 and len(chip) == 1:
        return verifyChIAPeaks(chia[0], chip[0], filename=filename, alpha=alpha)
    else:
        try:  # Check to see if multiple chiapet files are entered
            chia[0]
            inters = joinChIA(chia)
        except TypeError:
            inters = chia
        rankprod.performrankprod(chip, minentries=minentries, rankmethod=rank, alpha=alpha, filename=filename+'.bed')
        if threshold == 'binom':
            RP = bed.BedFile('T2_' + filename + '.bed', 'Peaks')

            return verifyChIAPeaks(inters, RP, filename=filename, alpha=alpha)
        elif threshold == 'alpha':
            RP = bed.BedFile('T1_' + filename + '.bed', 'Peaks')
            return verifyChIAPeaks(inters, RP, filename=filename, alpha=alpha)

        elif threshold == 'all':
            RP = bed.BedFile('ALL_' + filename + '.bed', 'Peaks')
            return verifyChIAPeaks(inters, RP, filename=filename, alpha=alpha)




from chipr import ival
import numpy as np
"""
Defines class for reading BED entries in the form of BED Paired End entries
"""


class BedPE:

    def __init__(self, chrom1, chromStart1, chromEnd1, chrom2, chromStart2, chromEnd2):
        self.chrom1 = chrom1
        self.chromStart1 = chromStart1
        self.chromEnd1 = chromEnd1
        self.chrom2 = chrom2
        self.chromStart2 = chromStart2
        self.chromEnd2 = chromEnd2
        self.blockCount = None
        self.usesstrand = False
        self.strand1 = None
        self.strand2 = None
        self.name1 = ''
        self.name2 = ''
        self.bounded = None
        self.partner1 = BedEntry(self.chrom1, self.chromStart1, self.chromEnd1)
        self.partner2 = BedEntry(self.chrom2, self.chromStart2, self.chromEnd2)

        if chrom1 == chrom2:
            self.loop = BedEntry(self.chrom1, self.chromEnd1, self.chromStart2)
            self.range = BedEntry(self.chrom1, self.chromStart1, self.chromEnd2)

        else:
            self.loop = -1
            self.range = -1

    def addOption(self,
                  name1=None,
                  name2=None,
                  score=None,
                  strand1=None,
                  strand2=None,
                  thickStart=None,
                  thickEnd=None,
                  itemRgb=None,
                  blockCount=None,
                  blockSizes=None,
                  blockStarts=None,
                  depth1=None,
                  depth2=None,
                  PETs=None,
                  signalValue=None,
                  pValue=None,
                  qValue=None,
                  peak=None,
                  tags=None,
                  summit1=None,
                  summit2=None,
                  fold=None,
                  fdr=None,
                  zscore=None,
                  bg=None, rank=None, comment=None):

        if name1: self.name = name1
        if name2: self.name = name2
        if score: self.score = score
        if strand1:
            self.strand1 = strand1
            self.usestrand = True  # use reverse complement when sequence is requested from genome
        if strand2:
            self.strand2 = strand2
            self.usestrand = True  # use reverse complement when sequence is requested from genome
        if thickStart: self.thickStart = thickStart
        if thickEnd: self.thickEnd = thickEnd
        if itemRgb: self.itemRgb = itemRgb
        if blockCount:
            self.blockCount = max(0, blockCount)
            if blockCount > 0:
                tblockSizes = [int(sizeword) for sizeword in blockSizes.rstrip(',').split(',')]
                tblockStarts = [int(startword) for startword in blockStarts.rstrip(',').split(',')]
                if len(tblockSizes) != blockCount or len(tblockStarts) != blockCount:
                    raise RuntimeError('Blockcount is incorrect in BED entry \"%s\"' % str(self))
                else:
                    self.blockSizes = tblockSizes
                    self.blockStarts = tblockStarts
        if depth1: self.depth1 = depth1
        if depth2: self.depth2 = depth2
        if PETs: self.PETs = PETs
        if signalValue: self.signalValue = signalValue
        if pValue: self.pValue = pValue
        if qValue: self.qValue = qValue
        if peak: self.peak = peak
        if tags: self.tags = tags
        if summit1: self.summit1 = summit1
        if summit2: self.summit2 = summit2
        if fold: self.fold = fold
        if fdr: self.fdr = fdr
        if bg: self.bg = bg
        if zscore: self.zscore = zscore
        if rank: self.rank = rank
        if comment: self.comment = comment

    def __str__(self):
        return str((self.chrom1, self.chromStart1, self.chromEnd1, self.chrom2, self.chromStart2, self.chromEnd2))

    def __getitem__(self, i):
        if self.blockCount:
            return (self.chrom1, self.chromStart1 + self.blockStarts[i],
                    self.chromStart1 + self.blockStarts[i] + self.blockSizes[i])

    def __iter__(self):
        if self.blockCount:
            for i in range(self.blockCount):
                if self.blockSizes[i] > 0:
                    yield (self.chrom1, self.chromStart1 + self.blockStarts[i],
                           self.chromStart1 + self.blockStarts[i] + self.blockSizes[i])

    def __len__(self):
        return self.blockCount

    def loc(self, genome=None, fixedwidth=None, usesummit=False, useshift=None):
        """ Retrieve the genomic location for BED entry, or sequence if genome is provided
            genome: a dictionary with keys for sequence names, e.g. 'chr1', 'chrX', etc, and values with indexed/sliceable strings
            fixedwidth: the width of the location/sequence if the width in the BED entry is ignored, and only its centre is used
            usesummit: centre a fixedwidth window around an assigned "summit"
            useshift: centre a fixedwidth window around a shifted centre point, e.g. useshift=-125 will shiftcentre point 125bp upstream,
            to say capture a fixedwidth=350bp window with 350/2-125=50bp downstream
        """
        otherstrand1 = False
        otherstrand2 = False
        if self.usesstrand:
            if self.strand1 == '-':
                otherstrand1 = True
            if self.strand2 == '-':
                otherstrand2 = True

        if otherstrand1 == False and otherstrand2 == False:
            end1 = self.chromEnd1
            start1 = self.chromStart1
            end2 = self.chromEnd2
            start2 = self.chromStart2
            mywidth1 = fixedwidth or (self.chromEnd1 - self.chromStart1)
            mycentre1 = start1 + (self.chromEnd1 - self.chromStart1) // 2
            mywidth2 = fixedwidth or (self.chromEnd2 - self.chromStart2)
            mycentre2 = start2 + (self.chromEnd2 - self.chromStart2) // 2
            if usesummit:
                mycentre1 = self.summit1
                mycentre2 = self.summit2
            if useshift:
                mycentre1 = mycentre1 + useshift
                mycentre2 = mycentre2 + useshift
            if fixedwidth:  # we need to re-calculate start and end
                if genome:
                    end1 = min(len(genome[self.chrom1]), mycentre1 + (mywidth1 // 2))
                    end2 = min(len(genome[self.chrom2]), mycentre2 + (mywidth2 // 2))
                else:
                    end1 = mycentre1 + (mywidth1 // 2)
                    end2 = mycentre2 + (mywidth2 // 2)
                start1 = max(0, mycentre1 - (mywidth1 // 2))
                start2 = max(0, mycentre2 - (mywidth2 // 2))

        #NOTE: TBC, need to add all strand configs
        else: # other strand
            start1 = self.chromEnd1
            end1 = self.chromStart1
            start2 = self.chromEnd2
            end2 = self.chromStart2
            mywidth = fixedwidth or (self.chromEnd1 - self.chromStart1)
            mycentre = self.chromStart1 + (self.chromEnd1 - self.chromStart1) // 2
            if usesummit:
                mycentre = self.summit
            if useshift:
                mycentre = mycentre - useshift  # shift is reversed on other strand
            if fixedwidth:  # we need to re-calculate start and end
                end = max(0, mycentre - (mywidth // 2))
                if genome:
                    start = min(len(genome[self.chrom1]), mycentre + (mywidth // 2))
                else:
                    start = mycentre + (mywidth // 2)

        if genome:  # refer to the genome sequence
            return [genome[self.chrom1][start1: end1], genome[self.chrom2][start2: end2]]
        else:
            return [self.chrom1, start1, end1, self.chrom2, start2, end2]

    def getInterval(self):
        return [ival.Interval(self.chromStart1, self.chromEnd1), ival.Interval(self.chromStart2, self.chromEnd2)]

    def getLength(self):
        return [self.chromEnd1 - self.chromStart1, self.chromEnd2 - self.chromStart2]

    def getDistance(self):
        if self.chrom1 == self.chrom2:
            d = abs(((self.chromStart1+self.chromEnd1)//2) - ((self.chromStart2+self.chromEnd2)//2))
            return d
        else:
            return None



"""
Defines class for reading in BED entries in a variety of formats
"""

class BedEntry:

    def __init__(self, chrom, chromStart, chromEnd):
        self.chrom = chrom
        self.chromStart = chromStart
        self.chromEnd = chromEnd
        self.blockCount = None
        self.usestrand = False
        self.strand = None
        self.name = ''

    def addOption(self,
                  name=None,
                  score=None,
                  strand=None,
                  thickStart=None,
                  thickEnd=None,
                  itemRgb=None,
                  blockCount=None,
                  blockSizes=None,
                  blockStarts=None,
                  strBlockSizes=None,
                  strBlockStarts=None,
                  signalValue=None,
                  pValue=None,
                  qValue=None,
                  peak=None,
                  localIDR=None,
                  globalIDR=None,
                  tags=None,
                  summit=None,
                  fold=None,
                  fdr=None,
                  zscore=None,
                  bg=None, rank=None,
                  rankProduct=None,
                  gene=None,
                  depth=None,
                  usestrand=False):
        if depth: self.depth = depth
        if name: self.name = name
        if score: self.score = score
        if strand:
            self.strand = strand
            self.usestrand = True # use reverse complement when sequence is requested from genome
        if usestrand: self.usestrand = usestrand
        if thickStart:
            if thickStart == 0:
                self.thickStart = "0"
            else:
                self.thickStart = thickStart
        if thickEnd: self.thickEnd = thickEnd
        if itemRgb: self.itemRgb = itemRgb
        if gene: self.gene = gene
        if blockCount:
            self.blockCount = max(0, blockCount)
            if blockCount > 0:
                tblockSizes = [int(sizeword) for sizeword in blockSizes.rstrip(',').split(',')]
                tblockStarts = [int(startword) for startword in blockStarts.rstrip(',').split(',')]
                if len(tblockSizes) != blockCount or len(tblockStarts) != blockCount:
                    raise RuntimeError('Blockcount is incorrect in BED entry \"%s\"' % str(self))
                else:
                    self.strBlockSizes = blockSizes
                    self.strBlockStarts = blockStarts
                    self.blockSizes = tblockSizes
                    self.blockStarts = tblockStarts

        if signalValue: self.signalValue = signalValue
        if pValue: self.pValue = pValue
        if qValue: self.qValue = qValue
        if peak: self.peak = peak
        if localIDR: self.localIDR = localIDR
        if globalIDR: self.globalIDR = globalIDR
        if tags: self.tags = tags
        if summit: self.summit = summit
        if fold: self.fold = fold
        if fdr: self.fdr = fdr
        if bg: self.bg = bg
        if zscore: self.zscore = zscore
        if rank: self.rank = rank
        if rankProduct: self.rankProduct = rankProduct

    def __str__(self):
        return str((self.chrom, self.chromStart, self.chromEnd))

    def __getitem__(self, i):
        if self.blockCount:
            return (self.chrom, self.chromStart + self.blockStarts[i], self.chromStart + self.blockStarts[i] + self.blockSizes[i])

    def __iter__(self):
        if self.blockCount:
            for i in range(self.blockCount):
                if self.blockSizes[i] > 0:
                    yield (self.chrom, self.chromStart + self.blockStarts[i], self.chromStart + self.blockStarts[i] + self.blockSizes[i])

    def __len__(self):
        return self.blockCount

    def loc(self, genome = None, fixedwidth = None, usesummit = False, useshift = None):
        """ Retrieve the genomic location for BED entry, or sequence if genome is provided
            genome: a dictionary with keys for sequence names, e.g. 'chr1', 'chrX', etc, and values with indexed/sliceable strings
            fixedwidth: the width of the location/sequence if the width in the BED entry is ignored, and only its centre is used
            usesummit: centre a fixedwidth window around an assigned "summit"
            useshift: centre a fixedwidth window around a shifted centre point, e.g. useshift=-125 will shiftcentre point 125bp upstream,
            to say capture a fixedwidth=350bp window with 350/2-125=50bp downstream
        """
        otherstrand = False
        if (self.usestrand):
            if (self.strand == '-'):
                otherstrand = True

        if (otherstrand == False):
            end = self.chromEnd
            start = self.chromStart
            mywidth = fixedwidth or (self.chromEnd - self.chromStart)
            mycentre = start + (self.chromEnd - self.chromStart) // 2
            if usesummit:
                mycentre = self.summit
            if useshift:
                mycentre = mycentre + useshift
            if fixedwidth: # we need to re-calculate start and end
                if genome:
                    end = min(len(genome[self.chrom]), mycentre + (mywidth // 2))
                else:
                    end = mycentre + (mywidth // 2)
                start = max(0, mycentre - (mywidth // 2))

        else: # other strand
            start = self.chromEnd
            end = self.chromStart
            mywidth = fixedwidth or (self.chromEnd - self.chromStart)
            mycentre = self.chromStart + (self.chromEnd - self.chromStart) // 2
            if usesummit:
                mycentre = self.summit
            if useshift:
                mycentre = mycentre - useshift # shift is reversed on other strand
            if fixedwidth: # we need to re-calculate start and end
                end = max(0, mycentre - (mywidth // 2))
                if genome:
                    start = min(len(genome[self.chrom]), mycentre + (mywidth // 2))
                else:
                    start = mycentre + (mywidth // 2)

        if genome: # refer to the genome sequence
            return genome[self.chrom][start : end]
        else:
            return (self.chrom, start, end)

    def setwidth(self, fixedwidth = None, usesummit = False):
        if fixedwidth:
            if usesummit:
                diff = self.summit - fixedwidth // 2
            else:
                diff = (self.chromEnd - self.chromStart) // 2 - fixedwidth // 2
            self.chromStart += diff
            self.chromStart += diff + fixedwidth
        return (self.chrom, self.chromStart, self.chromEnd)

    def getInterval(self):
        return ival.Interval(self.chromStart, self.chromEnd)

    def getLength(self):
        return self.chromEnd - self.chromStart




def _getLength(interval):
    return interval.chromEnd - interval.chromStart


def dist(entry1, entry2, signed = False, centre2centre = False):
    """ Calculate and return the BedEntry with the closest distance
        (from one end of the interval of this to the end of the interval of that).
        If centre2centre is True, use the centre-to-centre distance instead.
        If signed is True, the distance is negative if this interval is after the that.
    """
    if isinstance(entry1, BedEntry) and isinstance(entry2, BedEntry):
        if entry1.chrom == entry2.chrom:
            return ival.dist(entry1.getInterval(), entry2.getInterval(), signed, centre2centre)
    elif isinstance(entry1, BedPE) and isinstance(entry2, BedEntry):
        if entry1.chrom1 == entry1.chrom2 == entry2.chrom:
            return [ival.dist(entry1.getInterval()[0], entry2.getInterval(), signed, centre2centre),
                    ival.dist(entry1.getInterval()[1], entry2.getInterval(), signed, centre2centre)]
    elif isinstance(entry1, BedEntry) and isinstance(entry2, BedPE):
        if entry2.chrom1 == entry2.chrom2 == entry1.chrom:
            return [ival.dist(entry1.getInterval(), entry2.getInterval()[0], signed, centre2centre),
                    ival.dist(entry1.getInterval(), entry2.getInterval()[1], signed, centre2centre)]
    elif isinstance(entry1, BedPE) and isinstance(entry2, BedPE):
        if entry1.chrom1 == entry1.chrom2 == entry2.chrom1 == entry2.chrom2:
            return [ival.dist(entry1.getInterval()[0], entry2.getInterval()[0], signed, centre2centre),
                    ival.dist(entry1.getInterval()[0], entry2.getInterval()[1], signed, centre2centre),
                    ival.dist(entry1.getInterval()[1], entry2.getInterval()[0], signed, centre2centre),
                    ival.dist(entry1.getInterval()[1], entry2.getInterval()[1], signed, centre2centre)]
    return None


class BedFile:
    """ Read BED file.

        See http://genome.ucsc.edu/FAQ/FAQformat#format1

        The first three required BED fields are (part of all supported sub-formats):

        chrom - The name of the chromosome (e.g. chr3, chrY, chr2_random) or scaffold (e.g. scaffold10671).
        chromStart - The starting position of the feature in the chromosome or scaffold.
                     The first base in a chromosome is numbered 0.
        chromEnd - The ending position of the feature in the chromosome or scaffold.
                   The chromEnd base is not included in the display of the feature. .

        The 9 additional optional BED fields are (part of sub-format "Optional"):

        name - Defines the name of the BED line.
        score - A score between 0 and 1000.
                If the track line useScore attribute is set to 1 for this annotation data set,
                the score value will determine the level of gray in which this feature is displayed (higher numbers = darker gray).
        strand - Defines the strand - either '+' or '-'.
        thickStart - The starting position at which the feature is drawn thickly (for example, the start codon in gene displays).
        thickEnd - The ending position at which the feature is drawn thickly (for example, the stop codon in gene displays).
        itemRgb - An RGB value of the form R,G,B (e.g. 255,0,0).
                  If the track line itemRgb attribute is set to "On",
                  this RBG value will determine the display color of the data contained in this BED line.
        blockCount - The number of blocks (exons) in the BED line.
        blockSizes - A comma-separated list of the block sizes. The number of items in this list should correspond to blockCount.
        blockStarts - A comma-separated list of block starts. All of the blockStart positions should be calculated relative to chromStart.
                      The number of items in this list should correspond to blockCount.

        ENCODE also defines broadpeaks and narrowpeaks format (part of our "Peaks" sub-format):

        name - Defines the name of the BED line.
        score - Indicates how dark the peak will be displayed in the browser (0-1000).
        strand - +/- to denote strand or orientation (whenever applicable). Use '.' if no orientation is assigned.
        signalValue - Measurement of overall (usually, average) enrichment for the region.
        pValue - Measurement of statistical significance (-log10). Use -1 if no pValue is assigned.
        qValue - Measurement of statistical significance using false discovery rate (-log10). Use -1 if no qValue is assigned.
        peak - Point-source called for this peak; 0-based offset from chromStart. Use -1 if no point-source called.

        MACS also defines a "summit" peaks format (part of our "Summit" sub-format)
        It contains the peak summits locations for every peaks. The 5th column in this file is the .
        In addition to the required three, the following fields follow:
        length         [redundant, ignored]
        summit         summit height of fragment pileup
        tags
        pValue         [-10*log10(pvalue)]
        fold           [enrichment]
        FDR            [%; optional]

        "CCAT" BED-like file format:
        chromosome,
        peakcenter    [converted to summit],
        regionstart,
        regionend,
        tags          [tagcount],
        bg            [bgcount],
        zscore,
        fdr

    """

    def __init__(self, entries, format = 'Limited'):
        """
        Create a BedFile instance.
        :param entries: an iterable of entries or a filename
        :param format: the format of the BED file
        """
        self.format = format

        if isinstance(entries, str): # filename
            self.chroms = readBedFile(entries, format)
        else:
            self.chroms = dict()
            if format.lower().startswith('bedpe'):
                for entry in entries:
                    for num in range(1, 3):
                        if num == 1:
                            tree = self.chroms.get(entry.chrom1)
                            if not tree:
                                tree = ival.IntervalTree()
                                self.chroms[entry.chrom1] = tree
                            iv = ival.Interval(entry.chromStart1, entry.chromEnd1)
                            tree.put(iv, entry)

                        elif num == 2:
                            tree = self.chroms.get(entry.chrom2)
                            if not tree:
                                tree = ival.IntervalTree()
                                self.chroms[entry.chrom2] = tree
                                # put the entry in the interval tree for the appropriate chromosome
                                iv = ival.Interval(entry.chromStart2, entry.chromEnd2)
                                tree.put(iv, entry)
            else:
                for entry in entries:
                    # check if the chromosome has been seen before
                    tree = self.chroms.get(entry.chrom)
                    if not tree:
                        tree = ival.IntervalTree()
                        self.chroms[entry.chrom] = tree
                    # put the entry in the interval tree for the appropriate chromosome
                    iv = ival.Interval(entry.chromStart, entry.chromEnd)
                    tree.put(iv, entry)

    def __len__(self):
        n = 0
        for c in self.chroms:
            n += len(self.chroms[c])
        return n

    def generate(self, chrom):
        mytree = self.chroms.get(chrom)
        if mytree != None:
            for e in mytree:
                for entry in e.values:
                    yield entry

    def getChrom(self, chrom):
        mytree = self.chroms.get(chrom)
        entries = []
        if mytree != None:
            for e in mytree:
                for entry in e.values:
                    entries.append(entry)
        return BedFile(entries, self.format)

    def __iter__(self):
        self.chromqueue = ival.Stack()
        for c in sorted(self.chroms.keys())[::-1]:
            self.chromqueue.push(self.generate(c))
        self.current = self.chromqueue.pop()
        return self

    def __next__(self):
        try:
            ret = next(self.current)
        except StopIteration:
            if not self.chromqueue.isEmpty():
                self.current = self.chromqueue.pop()
                ret = next(self.current)
            else:
                raise StopIteration
        return ret

    def __contains__(self, item):
        if isinstance(item, BedEntry):
            tree = self.chroms.get(item.chrom)
            if tree is None:
                return False
            else:
                return ival.Interval(item.chromStart, item.chromEnd) in tree

        elif isinstance(item, BedPE):
            tree1 = self.chroms.get(item.chrom1)
            tree2 = self.chroms.get(item.chrom2)
            container = []
            if tree1 is None:
                container.append(False)
            else:
                container.append(ival.Interval(item.chromStart1, item.chromEnd1) in tree1)
            if tree2 is None:
                container.append(False)
            else:
                container.append(ival.Interval(item.chromStart2, item. chromEnd2) in tree2)
            return container
        else:
            return False

    def getOverlap(self, item):
        if isinstance(item, BedEntry):
            tree = self.chroms.get(item.chrom)
            if tree is None: return None
            else:
                iv = ival.Interval(item.chromStart, item.chromEnd)
                res = tree.isectall(iv)
                ret = []
                for r in res:
                    ret.extend(r.values)
                return ret
        elif isinstance(item, BedPE):
            tree1 = self.chroms.get(item.chrom1)
            tree2 = self.chroms.get(item.chrom2)
            container = []
            if tree1 is None:
                container.append(None)
            else:
                iv = ival.Interval(item.chromStart1, item.chromEnd1)
                res = tree1.isectall(iv)
                ret = []
                for r in res:
                    ret.extend(r.values)
                container.append(ret)
            if tree2 is None:
                container.append(None)
            else:
                iv = ival.Interval(item.chromStart2, item.chromEnd2)
                res = tree2.isectall(iv)
                ret = []
                for r in res:
                    ret.extend(r.values)
                container.append(ret)
            return container
        else:
            return None

    def getClosest(self, item):
        if isinstance(item, BedEntry):
            tree = self.chroms.get(item.chrom)
            if tree is None:
                return None
            else:
                iv = ival.Interval(item.chromStart, item.chromEnd)
                node = tree.closest(iv)
                if node is not None:
                    return node.values
                else:
                    return None
        elif isinstance(item, BedPE):
            tree1 = self.chroms.get(item.chrom1)
            tree2 = self.chroms.get(item.chrom2)
            container = []
            if tree1 is None:
                container.append(None)
            else:
                iv = ival.Interval(item.chromStart1, item.chromEnd1)
                node = tree1.closest(iv)
                if node is not None:
                    container.append(node.values)
                else:
                    container.append(None)
            if tree2 is None:
                container.append(None)
            else:
                iv = ival.Interval(item.chromStart1, item.chromEnd1)
                node = tree2.closest(iv)
                if node is not None:
                    container.append(node.values)
                else:
                    container.append(None)
            return container
        else:
            return None


    def getOneOfClosest(self, item):
        all = self.getClosest(item)
        if all == None: return None
        else: return next(iter(all))

    def getOneOfOverlap(self, item):
        all = self.getOverlap(item)
        if all == None: return None
        elif len(all) == 0: return None
        else: return next(iter(all))

    """
    getMetrics()
    Method for BedFile class that returns the mean, standard deviation, median, and median absolute deviation
    of the peak sizes for each chromosome of the bedfile
    """
    def getMetrics(self):
        chroms = self.chroms.keys()
        metrics = {}
        all_widths = []
        for chr in chroms:
            chromarray = self.generate(chr)
            entry = next(chromarray, 0)
            widthlist = []
            while entry != 0:
                widthlist.append(entry.chromEnd - entry.chromStart)
                all_widths.append(entry.chromEnd - entry.chromStart)
                entry = next(chromarray, 0)

            std = np.std(widthlist)
            widthmean = np.mean(widthlist)
            # abdiff = []
            m = np.median(widthlist)
            # Median Absolute Deviation is not necessary
            # for j in widthlist:
            #     abdiff.append(abs(j - m))
            # MAD = np.median(abdiff)

            metrics[chr] = [widthmean, std, m]
        # Stats for all peaks
        std = np.std(all_widths)
        widthmean = np.mean(all_widths)
        m = np.median(all_widths)
        metrics['total'] = [widthmean, std, m]

        return metrics

    def poolBED(self, bedfile):
        for entry in bedfile:
            # check if the chromosome has been seen before
            tree = self.chroms.get(entry.chrom)
            if not tree:
                tree = ival.IntervalTree()
                self.chroms[entry.chrom] = tree
            # put the entry in the interval tree for the appropriate chromosome
            iv = ival.Interval(entry.chromStart, entry.chromEnd)
            tree.put(iv, entry)




def readBedFile(filename, format = 'Limited'):
    """ Read a BED file.
        format: specifies the format of the file,
        "Limited", e.g.
            chr22 1000 5000
            chr22 2000 6000
        "Optional", e.g.
            track name=pairedReads description="Clone Paired Reads" useScore=1
            chr22 1000 5000 cloneA 960 + 1000 5000 0 2 567,488, 0,3512
            chr22 2000 6000 cloneB 900 - 2000 6000 0 2 433,399, 0,3601
            ...
            (also handles the Limited + score, and BED6 format)
        "Peaks", e.g.
            chr1    569780    569930    .    0    .    19    6.07811    -1    -1
            chr1    713300    713450    .    0    .    54    49.1167    -1    -1
        "Strand", e.g.
            chr4    185772359    185772424    -
            chr18    20513381    20513401    +
        also supports a 5th label field
            chr5    20611949        20611949        +       ENSG00000251629_20611949
            chr3    42187863        42187863        -       ENSG00000234562_42187863
        "Summit", e.g.
            # d = 130
            chr      start    end   length summit  tags -10*log10(pvalue)    fold_enrichment    FDR(%)
            chr1     8250     8671    422    286    46    145.84    11.68    0.51
            chr1    36382    36984    603    405    46    315.23    27.05    0.24
        "CCAT", e.g.
            chr8    94747805    94747070    94749250    525     3    21.519196    0.002000
            chr17   55277895    55277070    55279280    560    18    21.283333    0.002000
        "Cropped", e.g.
            chr1    851602    10
            chr1    921184    18
            chr1    931838    9
        "BedPE", e.g.
            chrom1  chromStart1 chromEnd1   chrom2  chromStart2 chromEnd2 + any number of additional fields
            chr1    85617       86100       chr1    120030      125039
            chr2    73891       74871       chr5    12709       12990
    """
    f = open(filename)
    row = 0
    acceptHeaderRows = 1
    headerRow = None
    sissrs = False
    gem = False
    start = False
    chroms = dict()
    for line in f:
        row += 1
        words = line.strip().split()
        if len(words) == 0:
            continue
        if words[0].strip().startswith('='):
            sissrs = True
            continue
        if words[0].strip().startswith('Position'):
            gem = True
            continue
        if sissrs:
            if words[0].strip().startswith('-'):
                start = True
                continue
            elif start:
                chrom = words[0]
                chromStart = int(words[1])
                chromEnd = int(words[2])
                entry = BedEntry(chrom, chromStart, chromEnd)
                entry.addOption(signalValue=int(words[3]), name = " ", score = int(words[3]), strand = '.',
                                pValue = float(-1), qValue = float(-1), peak = int(-1))

                # check if the chromosome has been seen before
                tree = chroms.get(chrom)
                if not tree:
                    tree = ival.IntervalTree()
                    chroms[chrom] = tree
                # put the entry in the interval tree for the appropriate chromosome
                iv = ival.Interval(entry.chromStart, entry.chromEnd)
                tree.put(iv, entry)
            else:
                continue
        elif gem:
            chrom, centre = words[0].split(':')
            centre = int(centre)
            chromStart = centre - 50
            chromEnd = centre + 50
            entry = BedEntry(chrom, chromStart, chromEnd)
            entry.addOption(signalValue=float(words[1]), name=" ", score=float(words[7]), strand=words[13], peak=centre,
                            pValue=float(words[6]), qValue=float(words[5]))
            tree = chroms.get(chrom)
            if not tree:
                tree = ival.IntervalTree()
                chroms[chrom] = tree
            # put the entry in the interval tree for the appropriate chromosome
            iv = ival.Interval(entry.chromStart, entry.chromEnd)
            tree.put(iv, entry)
        else:
            if len(words) == 0:
                continue # ignore empty lines
            if words[0].strip().startswith('#'):
                continue # comment
            if words[0].strip().startswith('browser'):
                continue # ignore
            if words[0].strip().startswith('track'):
                continue # ignore
            if words[1].strip().startswith('start'):
                continue # ignore
            try:
                if format.lower().startswith('bedpe'):
                    chrom1 = words[0]
                    chromStart1 = int(words[1])
                    chromEnd1 = int(words[2])
                    chrom2 = words[3]
                    chromStart2 = int(words[4])
                    chromEnd2 = int(words[5])

                    entry = BedPE(chrom1, chromStart1, chromEnd1, chrom2, chromStart2, chromEnd2)
                    if len(words) == 8:
                        entry.addOption(PETs=int(words[6]), pValue=float(words[7]))
                    if len(words) == 13:
                        entry.addOption(name1=words[6], name2=words[7], depth1=int(words[8]), depth2=int(words[9]),
                                        PETs=int(words[10]), pValue=float(words[11]), fdr=float(words[12]))
                    if chrom1 == chrom2:
                        tree = chroms.get(chrom1)
                        if not tree:
                            tree = ival.IntervalTree()
                            chroms[chrom1] = tree
                        iv1 = ival.Interval(entry.chromStart1, entry.chromEnd1)
                        iv2 = ival.Interval(entry.chromStart2, entry.chromEnd2)
                        tree.put(iv1, entry)
                        tree.put(iv2, entry)

                    else:
                        tree1 = chroms.get(chrom1)
                        tree2 = chroms.get(chrom2)
                        if not tree1:
                            tree1 = ival.IntervalTree()
                            chroms[chrom1] = tree1
                        if not tree2:
                            tree2 = ival.IntervalTree()
                            chroms[chrom2] = tree2
                        # put the entry in the interval tree for the appropriate chromosome
                        iv1 = ival.Interval(entry.chromStart1, entry.chromEnd1)
                        iv2 = ival.Interval(entry.chromStart2, entry.chromEnd2)
                        tree1.put(iv1, entry)
                        tree2.put(iv2, entry)
                else:
                    chrom = words[0]
                    if format.lower().startswith('ccat'):
                        chromStart = int(words[2])
                        chromEnd = int(words[3])
                    else: # all other standard BED formats
                        try:
                            chromStart = int(words[1])
                            chromEnd = int(words[2])
                        except ValueError:
                            print(words)
                            continue
                    entry = BedEntry(chrom, chromStart, chromEnd)
                    if format.lower().startswith('opt'):
                        if len(words) >= 9:
                            entry.addOption(name = words[3], score = float(words[4]), strand = words[5],
                                            thickStart = int(words[6]), thickEnd = int(words[7]), itemRgb = words[8])
                        elif len(words) >= 6:
                            entry.addOption(name = words[3], score = float(words[4]), strand = words[5])
                        elif len(words) >= 5:
                            entry.addOption(name = words[3], score = float(words[4]))
                        elif len(words) >= 4:
                            entry.addOption(name = words[3])
                        else:
                            entry.addOption(name = '.', score = int(words[3]), strand = '.')
                    elif format.lower().startswith('bed6'):
                        entry.addOption(name=words[3], score=float(words[4]), strand=words[5])
                    elif format.lower().startswith('strand'):
                        if len(words) >= 4: # properly formatted
                            entry.addOption(strand = words[3])
                        if len(words) >= 5:
                            entry.addOption(name = words[4])
                    elif format.lower().startswith('peak'):
                        if len(words) >= 10: # narrowpeaks
                            entry.addOption(name = words[3], score = int(words[4]), strand = words[5],
                                            signalValue = float(words[6]), pValue = float(words[7]),
                                            qValue = float(words[8]), peak = int(words[9]))
                        else: # broadpeaks
                            entry.addOption(name = words[3], score = int(words[4]), strand = words[5],
                                            signalValue = float(words[6]), pValue = float(words[7]),
                                            qValue = float(words[8]))
                    elif format.lower().startswith('rp'):
                        entry.addOption(name=words[3], score=int(words[4]), strand=words[5],
                                        signalValue=float(words[6]), pValue=float(words[7]),
                                        qValue=float(words[8]), rank=[float(r) for r in list(words[9].split(","))])
                    elif format.lower().startswith('idr'):
                        entry.addOption(name=words[3], score=int(words[4]), strand=words[5],
                                        signalValue=float(words[6]), pValue=float(words[7]),
                                        qValue=float(words[8]))
                    elif format.lower().startswith('2idr'):
                        #For IDR input with actual IDR values
                        entry.addOption(name=words[3], pValue=float(words[9]),
                                        qValue=float(words[10]))
                    elif format.lower().startswith('summit'):
                        if len(words) >= 9:
                            entry.addOption(summit = int(words[4]), tags = int(words[5]), pValue = float(words[6]),
                                            fold = float(words[7]), fdr = float(words[8]))
                        else:
                            entry.addOption(summit = int(words[4]), tags = int(words[5]),
                                            pValue = float(words[6]), fold = float(words[7]))
                    elif format.lower().startswith('ccat'):
                        entry.addOption(summit = int(words[1]) - entry.chromStart, tags = int(words[4]), bg = int(words[5]),
                                        zscore = float(words[6]), fdr = float(words[7]), name = '.',
                                        score = int(words[4]), strand = '.')
                    elif format.lower().startswith('crop'):
                        entry.addOption(score = int(words[2]), name = '.', strand = '.')
                        entry.chromEnd = entry.chromStart + 1
                    elif format.lower().startswith('bed12'):
                        entry.addOption(name=words[3], score=float(words[4]), strand=words[5], thickStart=int(words[6]),
                                        thickEnd=int(words[7]), itemRgb=words[8], blockCount=int(words[9]),
                                        blockSizes=words[10], blockStarts=words[11])
                    elif format.lower().startswith('TSS'):
                        entry.addOption(name=str(words[3]), gene=str(words[4]), strand=words[5])
                    elif format.lower().startswith('mspc'):
                        entry.addOption(name=str(words[3]), signalValue=float(words[4]))

                    # check if the chromosome has been seen before
                    tree = chroms.get(chrom)
                    if not tree:
                        tree = ival.IntervalTree()
                        chroms[chrom] = tree
                    # put the entry in the interval tree for the appropriate chromosome
                    iv = ival.Interval(entry.chromStart, entry.chromEnd)
                    tree.put(iv, entry)
            except RuntimeError as e:
                if not acceptHeaderRows:
                    raise RuntimeError('Error in BED file at row %d (%s)' % (row, e.strerror))
                else:
                    headerRow = words
                    acceptHeaderRows -= 1 # count down the number of header rows that can occur
    f.close()
    return chroms


'''
BEDPEtoBED12 and BED12toBEDPE
Set of helper functions to convert bed files between bedpe and BED12 format
'''


def BEDPEtoBED12(bedpe):
    bed12 = []
    for peEnt in bedpe:
        if peEnt.chrom1 == peEnt.chrom2:
            start = min(peEnt.chromStart1, peEnt.chromStart2)
            if start == 0:
                start = 1
            end = max(peEnt.chromEnd1, peEnt.chromEnd2)
            bed12Ent = BedEntry(peEnt.chrom1, start, end)
            bSizes = [peEnt.chromEnd1 - peEnt.chromStart1, peEnt.chromEnd2 - peEnt.chromStart2]
            bStarts = [0, max(peEnt.chromStart1, peEnt.chromStart2) - start]
            name = peEnt.chrom1+":"+str(peEnt.chromStart1)+".."+str(peEnt.chromEnd1)+"-"+peEnt.chrom2+":"\
                   +str(peEnt.chromStart2)+".."+str(peEnt.chromEnd2)
            bed12Ent.addOption(name=name, thickStart=int(start), thickEnd=int(end), strand=".", blockCount=2, score=500,
                               itemRgb="255,0,0", blockSizes=str(bSizes[0])+","+str(bSizes[1]),
                               blockStarts=str(bStarts[0])+","+str(bStarts[1]))
            bed12.append(bed12Ent)
        else:
            bed12Ent = BedEntry(peEnt.chrom1, peEnt.chromStart1, peEnt.chromEnd1)
            name = peEnt.chrom1 + ":" + str(peEnt.chromStart1) + ".." + str(peEnt.chromEnd1) + "-" + peEnt.chrom2 + ":" \
                    + str(peEnt.chromStart2) + ".." + str(peEnt.chromEnd2)
            bed12Ent.addOption(name=name, score=5, blockCount=1, thickStart=peEnt.chromStart1, thickEnd=peEnt.chromEnd1,
                               strand=".", itemRgb="255,0,0", blockSizes=str(peEnt.chromEnd1-peEnt.chromStart1),
                               blockStarts=str(0))
            bed12.append(bed12Ent)
    return bed12


def BED12toBEDPE(bed12):
    bedpe = []
    for ent in bed12:

        if ent.blockCount > 1:
            blockSizes = ent.blockSizes
            chrom1 = ent.chrom
            chrom2 = ent.chrom
            start1 = ent.chromStart
            end1 = start1+blockSizes[0]

            end2 = ent.chromEnd
            start2 = ent.chromEnd - blockSizes[1]
            peEnt = BedPE(chrom1, start1, end1, chrom2, start2, end2)
            bedpe.append(peEnt)
        else:
            if len(ent.name.split(':')) == 2:
                name = ent.name
                name = name.split('-')
                chrom1 = name.split(':')[0][0]
                chrom2 = name.split(':')[1][0]
                start1 = name.split(':')[0][1].split('..')[0]
                end1 = name.split(':')[0][1].split('..')[1]
                start2 = name.split(':')[1][1].split('..')[0]
                end2 = name.split(':')[1][1].split('..')[1]
                peEnt = BedPE(chrom1, start1, end1, chrom2, start2, end2)
                bedpe.append(peEnt)
    return bedpe


"""
Function that determines whether the entries of a BedPE file are bounded within the entries of another file.
Only useful when comparing BedPE files to TADs BedFiles
"""
def TADBoundary(BedFile, BedPE):
    boundedBedPE = []
    for ent in BedPE:
        if ent.loop is not None:
            searchChrom = BedFile.generate(ent.loop.chrom)
            for tad in searchChrom:
                if tad.chromStart <= ent.loop.chromStart and ent.loop.chromEnd <= tad.chromEnd:
                    ent.bounded = True
                    boundedBedPE.append(ent)
                    break
                else:
                    ent.bounded = False
    return boundedBedPE

def writeBedFile(entries, filename, format = 'BED6', header = None):
    """ Save the BED entries to a BED file.
        format - the format to use for WRITING
    """
    f = open(filename, 'w+')
    if header:
        f.write(header + '\n')
    for row in entries:
        if format == 'Peaks':
            f.write("%s\t%d\t%d\t%s\t%d\t%s\t%f\t%.16f\t%.16f" %
                    (row.chrom, row.chromStart, row.chromEnd, row.name,
                     row.score, row.strand, row.signalValue, row.pValue, row.qValue))
        elif format == 'RP':
            f.write("%s\t%d\t%d\t%s\t%d\t%s\t%f\t%.16f\t%.16f\t%s" %
                    (row.chrom, row.chromStart, row.chromEnd, row.name,
                     row.score, row.strand, row.signalValue, row.pValue, row.qValue, ','.join(str(r) for r in row.rank)))
        elif format == 'Limited':
            f.write("%s\t%d\t%d" % (row.chrom, row.chromStart, row.chromEnd))
        elif format == 'Strand':
            f.write("%s\t%d\t%d\t%s\t$s\t%s" % (row.chrom, row.chromStart, row.chromEnd, row.strand, row.name))
        elif format == 'BED12':
            f.write("%s\t%d\t%d\t%s\t%d\t%s\t%d\t%d\t%s\t%d\t%s\t%s"
                                % (row.chrom, row.chromStart, row.chromEnd,
                                    row.name, row.score, row.strand,
                                    row.thickStart, row.thickEnd, row.itemRgb,
                                    row.blockCount, row.strBlockSizes, row.strBlockStarts))
        elif format == 'BEDPE':
            f.write("%s\t%d\t%d\t%s\t%d\t%d\t%16f" % (row.chrom1, row.chromStart1, row.chromStart1,
                                                      row.chrom2, row.chromStart2, row.chromEnd2, row.pValue))
        elif format == 'contact1':
            f.write("%s\t%d\t%d\t%s\t%d\t%d\t%s\t%s" % (row.chrom1, (row.chromStart1+row.chromEnd1)//2,
                                                        (row.chromStart2+row.chromEnd2)//2,
                                                      row.chrom2, (row.chromStart1+row.chromEnd1)//2,
                                                        (row.chromStart2+row.chromEnd2)//2, '0,255,0', ' '))
        elif format == 'contact2':
            f.write("%s\t%d\t%d\t%s\t%d\t%d" % (row.chrom1, row.chromStart1, row.chromEnd1,
                                                      row.chrom2, row.chromStart2,  row.chromEnd2))
        else:
            f.write("%s\t%d\t%d\t%s\t%d\t%s" %
                    (row.chrom, row.chromStart, row.chromEnd, row.name, row.score, row.strand))
        f.write("\n")
    f.close()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 18 14:23:48 2017

@author: rhys newell
"""
#
from chipr import bed
from chipr import ival
import numpy as np
from functools import reduce
import operator
import math
import random
import copy
from chipr import multipletesting
import scipy.stats


def rankmetric(elem):
    return elem[1]


def orderentries(vec, reverse=True):
    return sorted(range(len(vec)), key=vec.__getitem__, reverse=reverse)


def prod(factors):
    return reduce(operator.mul, factors, 1)


def rankentries(vec, reverse=True, duphandling='average', random_seed=0.5):
    n = len(vec)
    vec1 = orderentries(vec, reverse=reverse)  # Ordered vector of entries
    vec2 = [vec[r] for r in vec1]  # List of ranks
    sumr = 0  # sum of ranks
    dups_amount = []
    duplicates = 0  # Count of duplicate ranks
    narray = [0] * n  # Empty array

    for i in range(n):
        sumr += i
        duplicates += 1

        if i == n - 1 or vec2[i] != vec2[i + 1]:
            dups_amount.append(duplicates)

            if duphandling == 'randomize':
                srank = i + 1 - duplicates
                rand_order = randomorder(srank, duplicates, random_seed=random_seed)
                it_rand = iter(rand_order)

                for j in range(i - duplicates + 1, i + 1):
                    narray[vec1[j]] = next(it_rand)
            elif duphandling =='average':
                averank = sumr / float(duplicates) + 1

                for j in range(i - duplicates + 1, i + 1):
                    narray[vec1[j]] = averank

            sumr = 0
            duplicates = 0

    return narray


def randomorder(i, n, random_seed=0.5):
    new_order = []

    for j in range(n):
        new_order.append(i + j)

    random.shuffle(new_order, lambda: random_seed)

    return new_order


def compare_with_ties(x, y):
    if x < y:
        return -1
    elif x > y:
        return 1
    else:
        return random.randint(0, 1) * 2 - 1


# Helper function that returns the overall distance of an interval
def intervaldist(entry):
    dist = entry.chromEnd - entry.chromStart
    return dist


def reduceEntries(bedf, metric, specifyMax):

    if metric.lower().startswith('signalvalue'):
        met = 'signalValue'
    elif metric.lower().startswith('pvalue'):
        met = 'pValue'
    elif metric.lower().startswith('qvalue'):
        met = 'qValue'
    else:
        met = 'signalValue'

    newbedf = []
    for i in bedf:
        ents = []
        sig = []
        for e in i:
            ents.append(e)
            sig.append(getattr(e, met))
        sortedents = [x for _, x in sorted(zip(sig, ents), key=lambda pair: pair[0], reverse=True)]
        if specifyMax is not None:
            b = bed.BedFile(sortedents[0:specifyMax], 'Peaks')
        else:
            b = bed.BedFile(sortedents, 'Peaks')
        newbedf.append(b)

    return newbedf


def rankBed(bedf, rankmethod='signalvalue'):
    '''
    Function for ranking entries for multiple replicates
    Entries are ranked independently between replicates

    @param bedf         a list of all replicates in bedfile format

    @param rankmethod   determines what metric is used to rank the entries
                        'all', 'signalvalue', 'pvalue', 'qvalue'
                        all: will only work when value sother than -1 are provided for each entry

    @param duphandling  determines how to handles duplicates
                        'randomize' randomizes the order of the ranks
                        'average' gives the average of ranks
    '''
    newbedf = []
    for rep in bedf:
        svals = []
        # rep = union(bf, minentries=1)[0]
        zcount = 0
        logged = False
        for peak in rep:
            try:
                if rankmethod.lower().startswith('signalvalue'):
                    svals.append(peak.signalValue)
                elif rankmethod.lower().startswith('pvalue'):
                    svals.append(peak.pValue)
                    if peak.pValue > 1:
                        logged = True
                elif rankmethod.lower().startswith('qvalue'):
                    if peak.qValue > 1:
                        logged = True
            except AttributeError:
                zcount += 1
                if logged:
                    svals.append(0)
                else:
                    svals.append(1)
        # zero_up = np.arange(len(rep) - zcount, len(rep), 1).tolist()
        # zmean = np.mean(zero_up)
        if max(svals) > 1:
            ranks = scipy.stats.rankdata(svals, method='average')
            max_rank = max(ranks)
            ranks = [max_rank - i for i in ranks]
        else:
            ranks = scipy.stats.rankdata(svals, method='average')
        for ent, rank in zip(rep, ranks):
            if ent != 0:
                ent.addOption(rank=rank)
            else:
                pass
        # newbed = bed.BedFile(sortedBed)
        newbedf.append(rep)
    return newbedf

def rankreps(bedf, minentries=None, rankmethod='signalvalue', duphandling='average', random_seed=0.5, specifyMax='maximum'):

    if specifyMax is None:
        specifyMax = 'maximum'
    elif isNumeric(specifyMax):
        specifyMax = int(specifyMax)
        print("Filtering BedFiles to have "+str(specifyMax)+" entries...")
        bedf = reduceEntries(bedf, rankmethod, specifyMax)
        bedf = list(iterflatten(bedf))
        print('Done!')
    elif specifyMax.lower().startswith('maximum'):
        bedf = list(iterflatten(bedf))
    elif specifyMax.lower().startswith('minimum'):

        lens = []
        for i in bedf:
            lens.append(len(i))

        specifyMax = min(lens)
        print("Filtering BedFiles to have " + str(specifyMax) + " entries...")
        bedf = reduceEntries(bedf, rankmethod, specifyMax)
        bedf = list(iterflatten(bedf))
        print('Done!')

    unions = union(bedf, minentries)
    rankedUni = rankBed(unions[1], rankmethod)
    rp = [1]*len(unions[0])

    for idx, rep in enumerate(rankedUni):
        zeros = rep.count(0)
        # Averages from n-zeros to n
        zero_up = np.arange(len(rep)-zeros, len(rep), 1).tolist()
        # All pseudo-peaks in all replicates receive rank n
        # zero_up = [len(rep)]*zeros
        zmean = np.mean(zero_up)
        if math.isnan(zmean):
            zmean = 1
        print(str(zeros) + ' Pseudo-peaks used for replicate number: ' + str(idx + 1))
        for i, ent in enumerate(rep):
            try:
                rp[i] *= ent.rank
            except AttributeError:
                rp[i] *= zmean

    return unions, rp, rankedUni


def rankEntropy(unions):
    """
    Entropy

    function for determining the entropy of the assigned ranks in an attempt to determine a 'surprisal' value

    @param unions - Directly takes the output from the rankreps function

    @return entropy_values - A list of entropy values for the for the replicates

    """
    entries = unions[1]
    entropy_values = []
    min_rank = 1
    max_rank = len(entries[0])
    for e in range(len(entries[0])):
        entry_ranks = []
        for rep in range(len(entries)):
            entry_ranks.append(entries[rep][e][1])
        if max(entry_ranks) - min(entry_ranks) != 0:
            z = [((i - min(entry_ranks))/(max(entry_ranks) - min(entry_ranks)))+1 for i in entry_ranks]
        else:
            z = [1 for i in entry_ranks]
        n = sum(z)
        # print(n)
        probs = [j/n for j in z]
        ent = -sum([k*math.log(k) for k in probs])
        entropy_values.append(ent)

    return entropy_values


def entropy(pvals):
    n = len(pvals)
    maxent = math.log2(n)
    vals = []
    sums = []
    # sumPs = 0
    for idx, i in enumerate(pvals):
        vi = i*math.log10(i)
        vals.append(vi)

        if -sum(vals) >= maxent:
            print(i)
            print(idx)
            break

    return vals, sums

"""Plan to remove this implementation. Causing peaks to spill over chromosomes boundaries"""
def extend(unions, mindist=100):
    print('Extending intervals')
    extended = []

    for ind, i in enumerate(unions[0]):
        dist = intervaldist(i[0])
        if dist < mindist:
            i[0].chromStart = i[0].chromStart - min(abs(i[1]-i[0].chromStart), 50)
            i[0].chromEnd = i[0].chromEnd + min(abs(i[2]-i[0].chromEnd), 50)
        extended.append(i[0])
    return extended, unions[1]


# Function to calculate the product of assigned ranks
# Finds product of "up" and "down" ranks
def product(vec):
    # products_down = []
    products_up = []
    k = len(vec)
    ranks = []
    for i in range(len(vec[0])):
        # ranks_down = []
        ranks_up = []
        for e in range(len(vec)):
            # ranks_down.append(vec[e][i][1])
            ranks_up.append(vec[e][i][1])
        # products_down.append(prod(ranks_down))
        products_up.append(prod(ranks_up))
        ranks.append(ranks_up)

    return products_up, ranks


# Function to calculate the monotonic transformation of the rank values
def monotonetransformation(vec):
    vals = []
    k = len(vec)
    logn = math.log(len(vec[0]) + 1, 10)

    for i in range(len(vec[0])):
        ranks = []
        for e in range(len(vec)):
            ranks.append(vec[e][i][1])
        logranks = [math.log(x / (len(vec[0]) + 1)) for x in ranks]
        vals.append(-(-k * logn + sum(logranks)))

    return vals


def rankprod(reps, method='all', holdout=None):
    '''
    rankprod

    Function for calculating RP values based on ranks applied to bed entries

    reps      array of bed entries and their union from the rankReps function
              If reps come from rankReps then input reps as reps[1]
              to avoid inputting the union list

    method    Specify whether to use all replicates for the product or holdout some
              e.g. 'all', 'holdout'

    holdout   specify how many replicates to withdraw from the analysis,
              removes replicates from the end of the list
    '''
    if method == 'all':
        obsproducts = product(reps[1])
        # RP = [x**(1/len(reps[1])) for x in obsproducts]
        if len(obsproducts[0]) == len(reps[0]):
            return reps, obsproducts[0], obsproducts[1]
        else:
            return [], obsproducts[0], obsproducts[1]

    # Perform rankprod on only some of the reps
    # The withheld reps can be compared against results to observe consistency
    # Automatically holds out items at end of the list of reps
    elif method == 'holdout':
        if holdout is None:
            holdout = 1
        k = len(reps[1])
        obsproducts = product(reps[1][0:(k - holdout)])

        if len(obsproducts[0]) == len(reps[0]):
            return reps, obsproducts[0], obsproducts[1]
        else:
            print("Holdout has resulted in uneven entry lists, please review your input data")


def isNumeric(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def switch(x):
    return {'a': 1, 'b': 2}.get(x, 9)


def iterflatten(root):  # Function flatten list of lists into single list, not limited by depth
    if isinstance(root, (list, tuple)):
        for element in root:
            for e in iterflatten(element):
                yield e
    else:
        yield root


def rankprodbounds(rho, n, k, Delta):
    '''
    rankprodbounds

    Description

    This function computes bounds on the p-value for rank products.


    Arguments

    rho     a vector of integers corresponding to the rank products for which one wishes to
            compute the p-value. These are NOT the geometric means of the rank product values.

    n       the number of peaks.

    k       the number of replicates.

    Delta   a character string indicating whether an upper bound ('upper'), lower bound
            ('lower'), or geometric approximation ('geometric') should be computed.


    Returns

    A vector of p-values, one for each rank product.

    Details

    The exact p-value is guaranteed to be in between the lower and the upper bound. The
    geometric mean of the two bounds can be used as an approximation. Each bound is a piecewise
    continuous function of the rank product. The different pieces each have an analytic form,
    the parameters of which can be computed recursively.

    Note

    This implementation closely follows the description in Heskes, Eisinga, Breitling:
    "A fast algorithm for determining bounds and accurate approximate p-values of the
    rank product statistic for replicate experiments", BMC Bioinformatics, referred to
    below to as HEB.
    '''
    # Input Handling
    if any(rho) > ((n ** k) + 1) or any(rho) < 1:
        return 'rho out of bounds'
    if isNumeric(Delta) is False:
        if Delta == 'geometric':
            temp1 = rankprodbounds(rho, n, k, 'upper')
            temp2 = rankprodbounds(rho, n, k, 'lower')
            # pexact = [np.sqrt(x * y) for x, y in zip(temp1[1], temp2[1])]
            pvalue = [np.sqrt(x * y) for x, y in zip(temp1, temp2)]  # geometric mean of upper and lower bound
            return pvalue
        elif Delta == 'upper':
            Delta = 1  # computing upper and lower bound
        elif Delta == 'lower':
            Delta = 0

    # Compute Intervals that contain the rank products
    logn = math.log(n)
    logrho = [math.log(x) for x in rho]
    logrhon = [-(y / logn) for y in logrho]
    allj = [math.ceil(z + k) for z in logrhon]  # Index specifying the interval that contains rho
    minj = min(allj)  # Lowest interval index
    maxj = max(allj)  # Highest interval index

    # Initialize Parameters
    param = np.empty(shape=[k + 1, maxj + 1], dtype=object)
    for i in range(k + 1):
        for j in range(int(maxj + 1)):
            param[i][j] = [[] for i in range(3)]
            param[i][j].append([0.0])
            param[i][j].append([0.0])
            # param is a matrix of lists; each element of param is a list with values for the parameters
    # 0 through 4, which correspond to the parameters alpha through epsilon in HEB;
    # specifically, param[i+1][j+1][0] corresponds to alpha_{i,j} in HEB, etc, where the offset
    # of 1 is introduced to be able to represent, for example, alpha_{0,0};
    # 0, 1, and 2 can be vectors (with possibly different lengths for different i and j),
    # 3 and 4 are scalars

    # COMPUTE PARAMETERS

    for j in range(minj, maxj + 1):
        param = updateparam(param, n, k, j, Delta)

    # call to the function updateparam which recursively computes all parameters that are needed
    # to calculate the p-value for a rank product rho that lies in the interval with index j

    # Compute Rank Products Given Parameters

    k1 = k + 1
    P = [0] * len(rho)
    G = [0] * len(rho)  # G is a vector of the same length as rho, for each rho bounding the number of rank products
    for j in range(int(minj), int(maxj + 1)):
        j1 = j + 1
        iii = [index for index, item in enumerate(allj) if item == j]  # Indices of all rps that fall in interval j
        thisrho = [rho[x] for x in iii]
        thisparam = copy.deepcopy(param[k1 - 1][j1 - 1])
        thisG = copy.deepcopy(thisparam[4])
        if j != 0:
            nrho = len(thisrho)
            nterms = len(thisparam[0])

            rhoparam = [i * thisparam[3][0] for i in thisrho]
            rhoparam = list(iterflatten(rhoparam))

            thisG = [thisG[0] + i for i in rhoparam]
            thisG = list(iterflatten(thisG))

            mat_thisparam2 = np.asarray(thisparam[2]).reshape((1, len(thisparam[2])))
            mat_thisrho = np.asarray(thisrho).reshape((1, len(thisrho)))

            # print(mat_thisparam2.shape)
            # print(mat_thisrho.shape)
            d1 = np.dot(mat_thisparam2.T, mat_thisrho)
            log_thisrho = []
            for x in thisrho:
                if x != 0:
                    log_thisrho.append(math.log(x))

            log_thisrho = np.asarray(log_thisrho).reshape((1, len(log_thisrho)))
            logn_thisparam1 = [logn * (k - j + x) for x in thisparam[1]]
            logn_thisparam1 = list(iterflatten(logn_thisparam1))

            mat_thisparam1 = np.asarray(logn_thisparam1).reshape((1, len(logn_thisparam1)))
            d2 = np.tile(log_thisrho, (nterms, 1)) - np.tile(mat_thisparam1, (nrho, 1)).T
            mat_thisparam0 = np.asarray(thisparam[0]).reshape((1, len(thisparam[0]))).T
            d3 = np.tile(mat_thisparam0, nrho)
            d2_expd3 = np.power(d2, d3)

            mats = np.multiply(d1, d2_expd3)
            mat_sum = np.sum(mats, axis=0)
            mat_sum = mat_sum.tolist()
            mat_sum = list(iterflatten(mat_sum))
            thisG = [x + y for x, y in zip(thisG, mat_sum)]

        it_thisG = iter(thisG)
        # it_matsum = iter(mat_sum)
        for i in iii:
            try:
                G[i] = next(it_thisG)
            except StopIteration:
                continue
            # P[i] = next(it_matsum)
        G = list(iterflatten(G))
        # P = list(iterflatten(P))

    # pexact = [i / (n ** k) for i in P]
    pvalue = [i / (n ** k) for i in G]

    return pvalue


def updateparam(param, n, k, j, Delta):
    '''

    updateparam

    Description

    This subroutine updates the current set of parameters to make sure that the parameters
    corresponding to k replicates and the j'th interval are included.

    Arguments

    param   a matrix of lists, where each element of param is a list with values for the
            parameters a through e; these parameters specify the functional form of the bound;
            a, b, and c are all vectors of unknown length, d and e are scalars.
    n       the number of peaks
    k       the number of replicates for which we need to compute the corresponding parameters.
    j       the index of the interval for which we need to compute the corresponding parameters.
    Delta   0 for the lower bound and 1 for the upper bound.

    Value

    A possibly updated set of parameters, at least including those corresponding to (k,j).

    Details

    This subroutine make sure that the parameters corresponding to k replicates and a rank product
    within the j'th interval are included. If they already are (because calculated before), it
    does not compute anything. Otherwise, it recursively computes all parameters
    that are needed to arrive at the parameters for (k,j).

    Note

    This implementation closely follows HEB, in particular equations (9) through (11).
    '''
    k1 = k + 1
    j1 = j + 1

    try:
        param[k1 - 1][j1 - 1]
    except IndexError:
        param = np.append(param, np.empty(shape=[1, j1], dtype=object), 0)  # empty, so needs calculating

        for i in range(j1 + 1):
            param[i][j1 - 1] = [[] for i in range(3)]
            param[i][j1 - 1].append([0.0])
            param[i][j1 - 1].append([0.0])

    if param[k1 - 1][j1 - 1][4][0] == 0.0:  # empty, so needs calculating
        if j == 0:
            param[k1 - 1][j1 - 1][4][0] = n ** k
            param[k1 - 1][j1 - 1][3][0] = 0.0
        else:
            k0 = k1 - 1
            j0 = j1 - 1
            param = updateparam(param, n, k - 1, j - 1, Delta)
            # checking that the parameters for (k-1,j-1) that are needed to compute the
            # parameters for (k,j) are indeed available; if not, they are themselves computed
            param00 = copy.deepcopy(param[k0 - 1][j0 - 1])
            try:
                newa0 = [x + 1 for x in param00[0]]
            except TypeError:
                newa0 = []
            newa0 = list(iterflatten(newa0))
            newb0 = copy.deepcopy(param00[1])
            newb0 = list(iterflatten(newb0))
            try:
                newc0 = [x / y for x, y in zip(param00[2], newa0)]
            except TypeError:
                newc0 = []
            newc0 = list(iterflatten(newc0))
            param11 = copy.deepcopy(param00)
            if k == j:  # Calculates param (a to e) when k = j
                param11[4] = [(1.0 - Delta) * (1.0 - x) for x in param00[4]]
                param11[4] = list(iterflatten(param11[4]))

                param11[3] = [Delta * x + y for x, y in zip(param00[3], param00[4])]
                param11[3] = list(iterflatten(param11[3]))

                param11[0] = [1.0, copy.deepcopy(param00[0]), copy.deepcopy(newa0)]
                param11[0] = list(iterflatten(param11[0]))

                param11[1] = [0.0, copy.deepcopy(param00[1]), copy.deepcopy(newb0)]
                param11[1] = list(iterflatten(param11[1]))

                Dparam002 = [Delta * x for x in param00[2]]
                param11[2] = [copy.deepcopy(param00[3]), copy.deepcopy(Dparam002), copy.deepcopy(newc0)]
                param11[2] = list(iterflatten(param11[2]))

            else:  # Calculates parameters (a to e)
                param = updateparam(param, n, k - 1, j, Delta)
                param01 = copy.deepcopy(param[k0 - 1][j1 - 1])

                logn = math.log(n)
                lognnkj = (k - j) * math.log(n)

                newa1 = [x + 1.0 for x in param01[0]]
                newa1 = list(iterflatten(newa1))

                newa = [copy.deepcopy(newa0), copy.deepcopy(newa1)]
                newa = list(iterflatten(newa))

                newb = [copy.deepcopy(newb0), copy.deepcopy(param01[1])]
                newb = list(iterflatten(newb))

                paramnewa = [-x / y for x, y in zip(param01[2], newa1)]
                newc = [copy.deepcopy(newc0), copy.deepcopy(paramnewa)]
                newc = list(iterflatten(newc))

                diffparam = [x - y for x, y in zip(param00[4], param01[4])]
                diffparam = list(iterflatten(diffparam))

                Dparam = [(Delta - 1) * x for x in diffparam]
                Dparam = list(iterflatten(Dparam))

                nparam = [n * param01[4][0]]
                param11[4] = [x + y for x, y in zip(nparam, Dparam)]
                param11[4] = list(iterflatten(param11[4]))

                lognminb = [[-1 * x * logn for x in param00[1]], [(1 - x) * logn for x in param01[1]]]
                lognminb = list(iterflatten(lognminb))

                logexp = [x ** y for x, y in zip(lognminb, newa)]
                logexp = list(iterflatten(logexp))

                newcxlog = [x * y for x, y in zip(newc, logexp)]
                newcxlog = list(iterflatten(newcxlog))

                paraDiv = (param00[4][0] - param01[4][0]) / np.exp(lognnkj)
                Dnparam013 = (1 - Delta) * param01[3][0] / n

                param11[3] = [Delta * x + Dnparam013 + paraDiv - sum(newcxlog) for x in param00[3]]
                param11[3] = list(iterflatten(param11[3]))

                param11[0] = [1.0, 1.0, copy.deepcopy(param00[0]), copy.deepcopy(param01[0]), copy.deepcopy(newa)]
                param11[0] = list(iterflatten(param11[0]))

                param11[1] = [0.0, 1.0, copy.deepcopy(param00[1]), copy.deepcopy(param01[1]), copy.deepcopy(newb)]
                param11[1] = list(iterflatten(param11[1]))

                Dparamc = [Delta * x for x in param00[2]]
                Dnewparamc = [(1 - Delta) * x / n for x in param01[2]]

                negparam013 = [-x for x in param01[3]]
                param11[2] = [copy.deepcopy(param00[3]), copy.deepcopy(negparam013), copy.deepcopy(Dparamc),
                              copy.deepcopy(Dnewparamc), copy.deepcopy(newc)]
                param11[2] = list(iterflatten(param11[2]))
            param[k1 - 1][j1 - 1] = makeunique(param11)

    return param


def makeunique(param):
    '''
    makeunique

    Description

    This subroutine updates the parameters for a specific number of replicates and interval
    such that it contains only unique combinations of the parameters a and b.

    Arguments

    param   a single list with values for the parameters a through e; these parameters
            specify the functional form of the bound; a, b, and c are all vectors of
            unknown length, d and e are scalars.

    Value

    A possibly updated and then more concise set of parameters containing only unique
    combinations of the parameters a and b.

    Details

    While updating the vectors a and b, one may end up with the exact same combinations of
    a and b. Given the functional form of the bound, the representation can then be made more
    concise by simply adding the corresponding elements of c.
    '''

    p1 = [x for x in param[0]]
    p2 = [x for x in param[1]]
    ab = list(zip(p1, p2))

    abunique = []  # The list of unique sets from pairing p1 and p2
    for i in ab:
        if i not in abunique:
            abunique.append(i)

    parr = np.array(abunique)
    ab = np.array(ab)
    nunique = len(abunique)
    param[0] = parr[:, 0].T.tolist()
    param[1] = parr[:, 1].T.tolist()
    newc = [0] * nunique

    for i in range(nunique):
        iii1 = np.where(ab[:, 0] == parr[i, 0])[0].tolist()
        iii2 = np.where(ab[:, 1] == parr[i, 1])[0].tolist()
        iii3 = list(set(iii1).intersection(iii2))
        l = []
        for j in iii3:
            l.append(param[2][j])
        newc[i] = sum(iterflatten(l))
    param[2] = newc

    return param


def matmult(a, b):

    zip_b = zip(*b)
    zip_b = list(zip_b)

    return [[sum(ele_a * ele_b for ele_a, ele_b in zip(row_a, col_b))
             for col_b in zip_b] for row_a in a]


def clip(interval, N, min_N, minsize=0):
    '''''
    clip

    Determining the set of intervals that define the N-fold intersection between the input intervals
    Note that zero-length intervals are NOT adding to the intersection count

    Interval: list of intervals

    '''''

    if min_N < 1:
        return "clip: N must be 1 or greater"
    ret = []
    mins = []
    maxs = []
    for iv in interval:
        mins.append(iv.min)
        maxs.append(iv.max)
    mins = sorted(mins)
    maxs = sorted(maxs)
    avmin = sum(mins)//len(mins)
    avmax = sum(maxs)//len(maxs)
    idx_min = 0
    idx_max = 0
    cnt = 0
    current = None

    while idx_min < len(mins) and idx_max < len(maxs):
        if mins[idx_min] < maxs[idx_max]:
            cnt += 1
            if current is None and cnt >= min_N:
                try:
                    if mins[idx_min+1] < maxs[idx_max]:
                        current = ival.Interval(mins[idx_min], mins[idx_min + 1])
                        if (current.max - current.min) >= minsize:
                            ret.append(current)
                    else:
                        current = ival.Interval(mins[idx_min], maxs[idx_max])
                        if (current.max - current.min) >= minsize:
                            ret.append(current)
                        current = None
                except IndexError:
                    current = ival.Interval(mins[idx_min], maxs[idx_max])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
                    current = None

            elif current is not None and cnt >= min_N:
                if current.max < mins[idx_min]:
                    current = ival.Interval(current.max, mins[idx_min])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
            idx_min += 1

        elif mins[idx_min] > maxs[idx_max]:

            if current is not None and cnt == min_N:
                if current.max < mins[idx_min]:
                    current = ival.Interval(current.max, mins[idx_min])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
                elif current.max < maxs[idx_max]:
                    current = ival.Interval(current.max, maxs[idx_max])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
            elif current is not None and cnt >= min_N:
                if current.max < mins[idx_min]:
                    current = ival.Interval(current.max, mins[idx_min])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
                elif current.max < maxs[idx_max]:
                    current = ival.Interval(current.max, maxs[idx_max])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
            cnt -= 1
            idx_max += 1

        else:
            if current is None and cnt >= min_N:
                try:
                    if mins[idx_min+1] < maxs[idx_max]:
                        current = ival.Interval(mins[idx_min], mins[idx_min + 1])
                        if (current.max - current.min) >= minsize:
                            ret.append(current)
                    else:
                        current = ival.Interval(mins[idx_min], maxs[idx_max])
                        if (current.max - current.min) >= minsize:
                            ret.append(current)
                        current = None
                except IndexError:
                    current = ival.Interval(mins[idx_min], maxs[idx_max])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
                    current = None


            elif current is not None and cnt >= min_N:
                if current.max < mins[idx_min]:
                    current = ival.Interval(current.max, mins[idx_min])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
            if current is not None and cnt == min_N:
                if current.max < mins[idx_min]:
                    current = ival.Interval(current.max, mins[idx_min])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
                elif current.max < maxs[idx_max]:
                    current = ival.Interval(current.max, maxs[idx_max])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
            elif current is not None and cnt >= min_N:
                if current.max < mins[idx_min]:
                    current = ival.Interval(current.max, mins[idx_min])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
                elif current.max < maxs[idx_max]:
                    current = ival.Interval(current.max, maxs[idx_max])
                    if (current.max - current.min) >= minsize:
                        ret.append(current)
            idx_max += 1
            idx_min += 1

    if idx_min < len(mins):
        return "Invalid intervals"
    if idx_max < len(maxs) and current is not None:
        while idx_max < len(maxs):
            if current.max < maxs[idx_max] and cnt >= min_N:
                current = ival.Interval(current.max, maxs[idx_max])
                if (current.max - current.min) >= minsize:
                    ret.append(current)
            cnt -= 1
            idx_max += 1
    newret = []
    for idx, r in enumerate(ret):
        if (r.max - r.min) >= minsize:
            newret.append(r)

    return newret, avmin, avmax


def merge(chrom, start, end, stash, min_N, N):
    '''
    merge

    Method for merging multiple entries into one,
    all falling within the given interval (which defines the extremes of the entries).
    With min_N set to 0 or 1, the method simply flattens the entries,
    resulting in exactly one entry with its score set to the maximum score (or number of entries, if score is not set).
    With a greater min_N, the method may return 0, 1 or more entries.

    chrom   the chromosome identifier

    start   the start point of the default interval to make up the entry

    end     the end point of the default interval to make up the entry

    stash   the entries to be merged

    min_N   minimum number of entries for defining an entry

    '''
    entries = []
    if min_N > len(stash):  # No point in the interval can meet the minimum number of overlapping entries
        # clipped = clip(stash, 1)
        return entries
    strand_aware = True  # Only false if one or more of the entries is false
    cnt_pos_strand = 0
    cnt_neg_strand = 0
    strand_reverse = False
    maxscore = 0
    for e in stash:
        if e.usestrand is False:
            strand_aware = False
        else:
            if e.usestrand is True:
                cnt_pos_strand += 1
                cnt_neg_strand += 0
            else:
                cnt_pos_strand += 0
                cnt_neg_strand += 1
    if strand_aware is True:
        strand_reverse = (cnt_neg_strand > cnt_pos_strand)
    nOnStrand = 0
    for e in stash:
        if strand_aware is True:
            if e.strand != strand_reverse:
                try:
                    maxscore = max(maxscore, e.score)
                except AttributeError:
                    maxscore = max(maxscore, e.signalValue)
                nOnStrand += 1
            else:
                try:
                    maxscore = max(maxscore, e.score)
                except AttributeError:
                    maxscore = max(maxscore, e.signalValue)
                nOnStrand += 1
    if min_N < 1:
        if strand_aware is True:
            entry = bed.BedEntry(chrom, start, end)
            entry.addOption(strand=strand_reverse, signalValue=stash[0].signalValue)
        else:
            entry = bed.BedEntry(chrom, start, end)
            entry.addOption(signalValue=stash[0].signalValue)
        if maxscore == 0:
            entry.addOption(score=nOnStrand)
        else:
            entry.addOption(score=maxscore)
        # clipped = clip(entry, 1)
        entries = entry
        return entries, entry.chromStart, entry.chromEnd
    else:
        ilist = []
        signalSum = 0
        depth = 0
        for support in stash:
            ilist.append(ival.Interval(support.chromStart, support.chromEnd))
            signalSum += support.signalValue
            depth += 1
        clipped = clip(ilist, N, min_N)
        iset = clipped[0]

        for iv in iset:

            if strand_aware:
                entry = bed.BedEntry(chrom, iv.min, iv.max)
                entry.strand = strand_reverse
                entry.addOption(signalValue=signalSum, depth=depth)
            else:
                entry = bed.BedEntry(chrom, iv.min, iv.max)
                entry.addOption(signalValue=signalSum, depth=depth)
            if maxscore == 0:
                entry.addOption(score=nOnStrand)
                entry.addOption(signalValue=signalSum, depth=depth)
            else:
                entry.addOption(score=maxscore)
                entry.addOption(signalValue=signalSum, depth=depth)
            entries.append(entry)
        return entries, clipped[1], clipped[2]


def compareTo(this_min, this_max, that_min, that_max):
    if this_min < that_min:
        return -1
    elif this_min > that_min:
        return 1
    elif this_max < that_max:
        return -1
    elif this_max > that_max:
        return 1
    else:
        return 0


def start(vec):

    """
    Return the minimum start point of a vector of entries
    """
    vec = [x for x in vec if x != 0 and x is not None]
    low = vec[0].chromStart
    startent = vec[0]
    for i in vec:
        if i.chromStart < low:
            low = i.chromStart
            startent = i
    return startent


def union(bedfiles, minentries=2, maxdist=2):
    '''
    union

    Create a set of BED entries that represent the union of BED entries in the given BED files.

    bedfiles     the BED files

    minentries   the minimum number of entries that need to fall into the interval to create a union

    maxdist      Specifies the maximum distance between entries before they are seen as no longer connected

    return the BED entries, to be used to construct a new BED file
    '''

    bedfiles = list(iterflatten(bedfiles))
    unions = []
    idx = -1
    maxzeros = len(bedfiles) - minentries
    if maxzeros < 0:
        maxzeros = 0
    entries = [[] for i in range(len(bedfiles))]
    edx = -1
    chroms = []

    for bedfile in bedfiles:
        for i in bedfile.chroms:
            if i not in chroms:
                chroms.append(i)

    for chrom in chroms:
        genarr = [0] * len(bedfiles)  # the generators, one for each chrom
        entryarr = [0] * len(bedfiles)  # the entry array, one for each chrom
        entryarr_prev = [0] * len(bedfiles)  # the entry array for the previous entries in entryarr
        stored = [0]*len(bedfiles)
        index = 0

        for bedfile in bedfiles:
            genarr[index] = bedfile.generate(chrom)
            try:
                entryarr[index] = next(genarr[index], 0)
            except StopIteration:
                print("no more chroms")
            index += 1

        for i in range(len(entryarr)):
            entryarr_prev[i] = entryarr[i]

        if 0 in entryarr:  # Account for mismatched chromosomes between bedfiles
            print("mismatch for " + str(chrom) + ", it has been removed.")
            if entryarr.count(0) > maxzeros:
                continue

        buildme = None
        indexs = []  # A history of all indexs of entryarr that appear in stashed
        stashed = []  # where we keep entries that will be merged, once entries are flattened into one
        while True:

            # index of current "head" BED entry
            # always look for the "first" entry in terms of their order
            first = -1

            for i in range(len(entryarr)):
                if entryarr[i] is not None and entryarr[i] != 0:
                    if first == -1:
                        first = i
                    else:
                        if compareTo(entryarr[first].chromStart, entryarr[first].chromEnd, entryarr[i].chromStart,
                                     entryarr[i].chromEnd) > 0:
                            first = i

            if first != -1:
                chosen = entryarr[first]
                if buildme is None:
                    buildme = ival.Interval(chosen.chromStart, chosen.chromEnd)
                    stashed.append(chosen)
                    indexs.append(first)
                else:  # Checks to see if buildme is complete
                    chosen_ival = ival.Interval(chosen.chromStart, chosen.chromEnd)
                    if ival.isect(buildme, chosen_ival):  # buildme is not complete, so continue extending
                        buildme = ival.Interval(buildme.min, max(buildme.max, chosen.chromEnd))
                        stashed.append(chosen)
                        indexs.append(first)

                    elif buildme.dist(chosen_ival, centre2centre=True) < maxdist:

                        # buildme not complete, because the next entry is not beyond the max distance
                        buildme = ival.Interval(buildme.min, max(buildme.max, chosen.chromEnd))
                        stashed.append(chosen)
                        indexs.append(first)

                    else:
                        # buildme is complete, stash the buildme interval (if sufficiently supported),
                        # then create a new buildme
                        s_ind = set(indexs)
                        if len(stashed) >= minentries and len(s_ind) >= minentries and (buildme.max - buildme.min) > 0:
                            merged = merge(chrom, buildme.min, buildme.max, stashed, minentries, N=len(bedfiles))
                            extentries = merged[0]
                            try:
                                for x in range(len(extentries)):
                                    if merged[0][x].chromEnd - merged[0][x].chromStart > 0:
                                        unions.append(merged[0][x])
                                        for i in range(len(entryarr_prev)):
                                            if entryarr_prev[i] in stashed:
                                                check = entryarr_prev[i].getInterval()
                                                mergcheck = merged[0][x].getInterval()
                                                if check.isectStrict(mergcheck):
                                                    stored[i] = copy.deepcopy(entryarr_prev[i])
                                                else:
                                                    stored[i] = 0

                                            elif entryarr[i] in stashed:
                                                check = entryarr[i].getInterval()
                                                mergcheck = merged[0][x].getInterval()
                                                if check.isectStrict(mergcheck):
                                                    stored[i] = copy.deepcopy(entryarr[i])
                                                else:
                                                    stored[i] = 0

                                            else:
                                                stored[i] = 0
                                        for i in range(len(stored)):
                                            entries[i].append(stored[i])

                            except TypeError:
                                if merged[0].chromEnd - merged[0].chromStart > 0:

                                    unions.append(merged[0])
                                    for i in range(len(entryarr_prev)):
                                        if entryarr_prev[i] in stashed:
                                            check = entryarr_prev[i].getInterval()
                                            mergcheck = merged[0].getInterval()
                                            if check.isectStrict(mergcheck):
                                                stored[i] = copy.deepcopy(entryarr_prev[i])
                                            else:
                                                stored[i] = 0
                                        elif entryarr[i] in stashed:
                                            check = entryarr[i].getInterval()
                                            mergcheck = merged[0].getInterval()
                                            if check.isectStrict(mergcheck):
                                                stored[i] = copy.deepcopy(entryarr[i])
                                            else:
                                                stored[i] = 0
                                        else:
                                            stored[i] = 0
                                    for i in range(len(stored)):
                                        entries[i].append(stored[i])


                            idx += 1
                        buildme = ival.Interval(chosen.chromStart, chosen.chromEnd)
                        stashed = []
                        stashed.append(chosen)
                        indexs = []
                        indexs.append(first)
                entryarr_prev[first] = entryarr[first]
                entryarr[first] = next(genarr[first], 0)

            if first == -1:
                break

        # there is probably a final entry, which also needs to be stashed away
        s_ind = set(indexs)
        if buildme is not None and len(stashed) >= minentries and len(s_ind) >= minentries and (buildme.max - buildme.min) > 0:
            merged = merge(chrom, buildme.min, buildme.max, stashed, minentries, N=len(bedfiles))
            extentries = merged[0]
            try:
                for x in range(len(extentries)):
                    if merged[0][x].chromEnd - merged[0][x].chromStart > 0:

                        unions.append(merged[0][x])
                        for i in range(len(entryarr_prev)):
                            if entryarr_prev[i] in stashed:
                                check = entryarr_prev[i].getInterval()
                                mergcheck = merged[0][x].getInterval()
                                if check.isectStrict(mergcheck):
                                    stored[i] = copy.deepcopy(entryarr_prev[i])
                                else:
                                    stored[i] = 0

                            elif entryarr[i] in stashed:
                                check = entryarr[i].getInterval()
                                mergcheck = merged[0][x].getInterval()
                                if check.isectStrict(mergcheck):
                                    stored[i] = copy.deepcopy(entryarr[i])
                                else:
                                    stored[i] = 0

                            else:
                                stored[i] = 0
                        for i in range(len(stored)):
                            entries[i].append(stored[i])

            except TypeError:
                if merged[0].chromEnd - merged[0].chromStart > 0:

                    unions.append(merged[0])
                    for i in range(len(entryarr_prev)):
                        if entryarr_prev[i] in stashed:
                            check = entryarr_prev[i].getInterval()
                            mergcheck = merged[0].getInterval()
                            if check.isectStrict(mergcheck):
                                stored[i] = copy.deepcopy(entryarr_prev[i])
                            else:
                                stored[i] = 0
                        elif entryarr[i] in stashed:
                            check = entryarr[i].getInterval()
                            mergcheck = merged[0].getInterval()
                            if check.isectStrict(mergcheck):
                                stored[i] = copy.deepcopy(entryarr[i])
                            else:
                                stored[i] = 0
                        else:
                            stored[i] = 0
                    for i in range(len(stored)):
                        entries[i].append(stored[i])

            idx += 1

    return unions, entries


def overlaps(ent1, ent2, buffer=0):
    """
    Overlaps:
    defines whether two bed entries are overlapping with each other
    @param buffer: sets a buffer zone around entry1 extending it's space by that many bps
    """
    if (ent1.chromStart-buffer) <= ent2.chromStart <= (ent1.chromEnd+buffer):
        return True
    elif (ent1.chromStart-buffer) <= ent2.chromEnd <= (ent1.chromEnd+buffer):
        return True
    else:
        return False


def collapse(bedf, idr=False):
    """
    collapse:
    Function that collapses the final union bedfile AFTER the rank product analysis has been performed. If there are multiple
    Bed Entries overlapping one another, they will be merged preventing an overabundance of bed entries created by the
    union function.
    """
    allOl = []
    pvals = []
    if idr is True:
        for chrom in bedf.chroms:
            chromarr = bedf.generate(chrom)
            ol1 = []
            ol2 = []
            ol3 = []
            ol4 = []
            ol5 = []

            for entry in chromarr:

                if len(ol3) == 0:
                    cnt = 1
                    ol1.append(entry.chromStart)
                    ol2.append(entry.chromEnd)
                    ol3.append(entry)
                    try:
                        ol4.append(entry.score)
                    except AttributeError:
                        ol4.append(0)
                    ol5.append(entry.signalValue)

                else:
                    for o in ol3:
                        if overlaps(o, entry):
                            ol1.append(entry.chromStart)
                            ol2.append(entry.chromEnd)
                            ol3.append(entry)
                            try:
                                ol4.append(entry.score)
                            except AttributeError:
                                ol4.append(0)
                            ol5.append(entry.signalValue)

                            break
                        else:
                            cnt += 1

                if cnt > len(ol3):
                    if len(ol3) > 1:
                        mer = bed.BedEntry(chrom, min(ol1), max(ol2))
                        newScore = max(ol4)  # Calculate combined p-value
                        newSignal = sum(ol5)
                        mer.addOption(name='TBD',
                                      score=newScore,
                                      strand='.',
                                      signalValue=newSignal,
                                      pValue=-1,
                                      qValue=-1)
                        allOl.append(mer)

                    else:
                        ol3[0].addOption(name='TBD',
                                         score=max(ol4),
                                         strand='.',
                                         signalValue=sum(ol5),
                                         pValue=-1,
                                         qValue=-1)
                        allOl.append(ol3[0])

                    ol1 = []
                    ol2 = []
                    ol3 = []
                    ol4 = []
                    ol5 = []

                    cnt = 1
                    ol1.append(entry.chromStart)
                    ol2.append(entry.chromEnd)
                    ol3.append(entry)
                    try:
                        ol4.append(entry.score)
                    except AttributeError:
                        ol4.append(0)
                    ol5.append(entry.signalValue)

        return allOl
    else:
        for chrom in bedf.chroms:
            chromarr = bedf.generate(chrom)
            ol1 = []
            ol2 = []
            ol3 = []
            ol4 = []
            ol5 = []
            ol6 = []

            for entry in chromarr:

                if len(ol3) == 0:
                    cnt = 1

                    ol1.append(entry.chromStart)
                    ol2.append(entry.chromEnd)
                    ol3.append(entry)
                    try:
                        ol4.append(entry.pValue)
                    except AttributeError:
                        ol4.append(-1)
                    try:
                        ol5.append(entry.qValue)
                    except AttributeError:
                        ol5.append(-1)
                    try:
                        ol6.append(entry.signalValue)
                    except AttributeError:
                        ol6.append(-1)
                else:
                    for o in ol3:
                        if overlaps(o, entry):
                            ol1.append(entry.chromStart)
                            ol2.append(entry.chromEnd)
                            ol3.append(entry)
                            try:
                                ol4.append(entry.pValue)
                            except AttributeError:
                                ol4.append(-1)
                            try:
                                ol5.append(entry.qValue)
                            except AttributeError:
                                ol5.append(-1)
                            try:
                                ol6.append(entry.signalValue)
                            except AttributeError:
                                ol6.append(-1)
                            break
                        else:
                            cnt += 1
                if cnt > len(ol3):
                    if len(ol3) > 1:
                        mer = bed.BedEntry(chrom, min(ol1), max(ol2))
                        newp = scipy.stats.combine_pvalues(ol4, 'stouffer')[1] #Calculate combined p-value
                        pvals.append(newp)
                        newq = scipy.stats.combine_pvalues(ol5, 'stouffer')[1]
                        # qvals.append(newq)

                        mer.addOption(name='TBD',
                                      score=min([abs(int(125 * math.log2(newp))), 1000]),
                                      strand='.',
                                      signalValue=sum(ol6),
                                      pValue=newp,
                                      qValue=newq)
                        allOl.append(mer)

                    else:
                        try:
                            ol3[0].addOption(name='TBD',
                                             score=min([abs(int(125 * math.log2(ol4[0]))), 1000]),
                                             strand='.',
                                             signalValue=sum(ol6),
                                             pValue=ol4[0],
                                             qValue=ol5[0])
                            allOl.append(ol3[0])
                            pvals.append(ol4[0])
                        except ValueError:
                            ol3[0].addOption(name='TBD',
                                             score=min([abs(int(125 * math.log2(1))), 1000]),
                                             strand='.',
                                             signalValue=sum(ol6),
                                             pValue=-1,
                                             qValue=-1)
                            allOl.append(ol3[0])
                            pvals.append(-1)
                    ol1 = []
                    ol2 = []
                    ol3 = []
                    ol4 = []
                    ol5 = []
                    ol6 = []
                    cnt = 1
                    ol1.append(entry.chromStart)
                    ol2.append(entry.chromEnd)
                    ol3.append(entry)
                    try:
                        ol4.append(entry.pValue)
                    except AttributeError:
                        ol4.append(-1)
                    try:
                        ol5.append(entry.qValue)
                    except AttributeError:
                        ol5.append(-1)
                    try:
                        ol6.append(entry.signalValue)
                    except AttributeError:
                        ol6.append(-1)

        return allOl, pvals


def thresholdCalc(p, k=6):

    """
    thresholdCalc
    Function for calculating a suitable cutoff threshold for a list of P-values.
    Finds intersection point between the binomial cumulative distribution function and the rank product probabilities

    params:
    @p - A list of P-values from the rank product distribution

    """
    n = len(p)
    ps = sorted(p)
    bs = []
    for i, v in enumerate(ps):
        b = scipy.stats.binom.cdf(i, n, v)
        bs.append(b)

    idx = np.argwhere(np.diff(np.sign(np.array(ps) - np.array(bs))) != 0).reshape(-1) + 0
    pits = [ps[i] for i in idx]

    return bs, ps, pits, idx


def connect_entries(bedf, reps, default_min_peak=20, broadpeaks=False):
    """
    connect_entries
    Modified version of union function that takes only one bedfile input and connects entries that are within 1 base pair
    of each other and have a significantly close enough rank product p-value.

    Prevents over fragmentation of peaks.
    """
    chroms = bedf.chroms.keys()

    unions = []
    for chrom in chroms:
        genarr = bedf.generate(chrom)
        entryarr = next(genarr, 0)
        buildme = None
        indexs = []  # A history of all indexs of entryarr that appear in stashed
        stashed = []  # where we keep entries that will be merged, once entries are flattened into one
        while True:

            # index of current "head" BED entry
            # always look for the "first" entry in terms of their order
            if entryarr != 0:
                chosen = entryarr
            if buildme is None:

                buildme = ival.Interval(chosen.chromStart, chosen.chromEnd)
                buildme_vars = vars(chosen)

                stashed.append(chosen)

            else:  # Checks to see if buildme is complete
                chosen_ival = ival.Interval(chosen.chromStart, chosen.chromEnd)
                chosen_vars = chosen.pValue

                if buildme.dist(chosen_ival, centre2centre=False) <= 1:
                    # Test if either interval is shorter than expected
                    lowerbound = calcLowerBound(reps, chosen, default_min_peak, broadpeaks)

                    if lowerbound >= chosen_ival.max - chosen_ival.min \
                            or lowerbound >= buildme.max - buildme.min:
                        # If so, test if they should be joined
                        if round(buildme_vars['pValue'], 1) == round(chosen_vars, 1):  # Test if peaks are similar
                            buildme = ival.Interval(buildme.min, max(buildme.max, chosen.chromEnd))
                            if buildme_vars['pValue'] > chosen_vars:
                                buildme_vars = vars(chosen)
                            stashed.append(chosen)
                        else:  # Treat peaks as independent entities

                            # buildme is complete, stash the buildme interval (if sufficiently supported),
                            # then create a new buildme
                            lowerbound = calcLowerBound(reps, chosen, default_min_peak, broadpeaks)

                            if lowerbound <= buildme.max - buildme.min:
                                merged = bed.BedEntry(chrom, buildme.min, buildme.max)
                                del buildme_vars['chrom'], buildme_vars['chromStart'], buildme_vars['chromEnd']
                                merged.addOption(**buildme_vars)
                                unions.append(merged)

                            buildme = ival.Interval(chosen.chromStart, chosen.chromEnd)
                            buildme_vars = vars(chosen)

                    elif round(buildme_vars['pValue'], 3) == round(chosen_vars, 3): # Test if peaks are similar

                        buildme = ival.Interval(buildme.min, max(buildme.max, chosen.chromEnd))
                        if buildme_vars['pValue'] > chosen_vars:
                            buildme_vars = vars(chosen)
                        stashed.append(chosen)
                        # entryarr = next(genarr, 0)
                    else: # Treat peaks as independent entities

                        # buildme is complete, stash the buildme interval (if sufficiently supported),
                        # then create a new buildme
                        lowerbound = calcLowerBound(reps, chosen, default_min_peak, broadpeaks)

                        if lowerbound <= buildme.max - buildme.min:
                            merged = bed.BedEntry(chrom, buildme.min, buildme.max)
                            del buildme_vars['chrom'], buildme_vars['chromStart'], buildme_vars['chromEnd']
                            merged.addOption(**buildme_vars)
                            unions.append(merged)

                        buildme = ival.Interval(chosen.chromStart, chosen.chromEnd)
                        buildme_vars = vars(chosen)

                else:

                    # buildme is complete, stash the buildme interval (if sufficiently supported),
                    # then create a new buildme
                    lowerbound = calcLowerBound(reps, chosen, default_min_peak, broadpeaks)

                    if lowerbound <= buildme.max - buildme.min:
                        merged = bed.BedEntry(chrom, buildme.min, buildme.max)
                        del buildme_vars['chrom'], buildme_vars['chromStart'], buildme_vars['chromEnd']
                        merged.addOption(**buildme_vars)
                        unions.append(merged)

                    buildme = ival.Interval(chosen.chromStart, chosen.chromEnd)
                    buildme_vars = vars(chosen)

                    stashed = []
                    stashed.append(chosen)

                # entryarr_prev = entryarr
                entryarr = next(genarr, 0)

                if entryarr == 0:
                    break

        # there is probably a final entry, which also needs to be stashed away
        s_ind = set(indexs)
        if buildme is not None:
            lowerbound = calcLowerBound(reps, chosen, default_min_peak, broadpeaks)

            if lowerbound <= buildme.max - buildme.min:
                merged = bed.BedEntry(chrom, buildme.min, buildme.max)
                del buildme_vars['chrom'], buildme_vars['chromStart'], buildme_vars['chromEnd']
                merged.addOption(**buildme_vars)
                unions.append(merged)

    unions = bed.BedFile(unions, 'IDR')

    return unions


def calcLowerBound(reps, chosen, default_min_peak=20, broadpeaks=False):
    closest_widths = []
    for rep in reps:
        closest_peaks = rep.getClosest(chosen)
        if closest_peaks is not None:
            for closest in closest_peaks:
                closest_widths.append(closest.chromEnd - closest.chromStart)
    if len(closest_widths) != 0 and not broadpeaks:
        mean_width = np.mean(closest_widths)
        lowerbound = mean_width - mean_width / 2
        upperbound = mean_width + mean_width / 2
    elif broadpeaks:
        lowerbound = default_min_peak
    else:
        lowerbound = default_min_peak
        upperbound = 450

    return lowerbound


def performrankprod(bedf, minentries=2, rankmethod="signalvalue", specifyMax=None,
                    duphandling='average', random_seed=0.5,
                    alpha=0.05,
                    filename="bedfile_unions.bed",
                    default_min_peak=20,
                    print_pvals=True,
                    fragment=False):
    print("Using high fragmentation:", fragment)

    # First create intersection and rank the entries in each replicate and return the rankproduct values
    ranks = rankreps(bedf, minentries, rankmethod, duphandling, random_seed, specifyMax)

    # Calculate rank product for each entry that contributes to a union entry
    # Calculate the pvalues of the rank product values
    print('Calculating rank product probabilities...')
    rpb_up = rankprodbounds(ranks[1], len(ranks[1]), len(bedf), 'geometric')
    print('Calculating binomial threshold...')
    # Calculate rpb and binomial intersection point
    Pks = thresholdCalc(rpb_up, k=len(bedf)-(minentries-1))
    if len(Pks[2]) != 0:
        binomAlpha = round(min(Pks[2]), 3)
    else:
        print('No binomial convergence, defaulting to 0.1')
        binomAlpha = 0.1

    # Perform multiple hypothesis testing correction upon the pvals
    fdr = multipletesting.fdrcorrection(rpb_up)

    # Determine whether to remove entries that are called significant
    print('Cleaning up output...')
    for i, v in enumerate(ranks[0][0]):
        p = rpb_up[i]

        if p != 0.0:
            ranks[0][0][i].addOption(name='TBD',
                                     score=min([abs(int(125 * math.log2(rpb_up[i]))), 1000]),
                                     strand='.',
                                     pValue=rpb_up[i],
                                     qValue=fdr[1][i])
        else:
            ranks[0][0][i].addOption(name='TBD',
                                     score=1000,
                                     strand='.',
                                     pValue=2.5e-20,
                                     qValue=2.5e-20)
    collapsed = bed.BedFile(ranks[0][0], 'IDR')
    len1 = len(collapsed)

    collapsed = connect_entries(collapsed, bedf, default_min_peak, fragment)

    t3cnt = 0
    t1_unions = []
    # t2_unions = []
    t3_unions = []
    pvals = []

    for i, v in enumerate(collapsed):
        """
        Determine what tier each union falls under
        Tier 1 - Union with significance score <= alpha ~~~~~~ OLD ~~~~~~~

        primary - Union with significance score <= binomAlpha.
                 These are unions that the binomial threshold calculation has deemed to be significant.

        secondary - Union that does not meet requirements for previous Tiers.
                 Should not be discarded as its peaks still appear in majority of replicates.
        """
        pvals.append(v.pValue)
        if v.pValue <= binomAlpha:
            v.addOption(name='primary_peak_' + str(i),
                        strand='.')
            t1_unions.append(v)
        # elif v.pValue <= binomAlpha:
        #     v.addOption(name='T2_peak_' + str(i),
        #                 strand='.')
        #     t2_unions.append(v)
        else:
            t3cnt += 1
            v.addOption(name='secondary_peak_' + str(i),
                        strand='.')
            t3_unions.append(v)

    sortedUnions = [x for _, x in sorted(zip(pvals, collapsed), key = lambda pair:pair[0])]
    print(round((len(t1_unions)/len(collapsed))*100, 2), "% Primary peaks")
    print(round((t3cnt/len(collapsed))*100, 2), "% Secondary peaks")

    if filename is not None:
        bed.writeBedFile(sortedUnions, filename + "_all.bed", format="Peaks")
        bed.writeBedFile(t1_unions, filename + "_optimal.bed", format="Peaks")
        mets = bed.BedFile(sortedUnions).getMetrics()
        with open(filename+'_log.txt', 'w') as f:
            f.write('Primary Peaks: '+str(len(t1_unions))+'\n'+
                    'Secondary Peaks: '+str(t3cnt)+'\n'+
                    'Total Peaks:'+str(t3cnt+len(t1_unions))+'\n')
            f.write('\n Chromosome'+'\t mean peak size\t standard deviation: \n')
            for k, v in mets.items():
                f.write(k+'\t'+str(v[0])+'\t'+str(v[1])+'\n')

    return collapsed, Pks, rpb_up, fdr


import unittest
from chipr import bed
from chipr.rankprod import *

class TestRankProdMethods(unittest.TestCase):

    def test_bed(self):
        bf = bed.BedFile("test/data/med/med1_peaks.broadPeak", "Peaks")
        print(bf.chroms.keys())
        g = bf.generate('chr1')
        print(next(g))
        print(next(g))
        print(next(g))
        cnt = 0
        for entry in bf:
            cnt += 1
            print(str(cnt) + '\t' + str(entry))
            if cnt == 100:
                break
        entry1 = bed.BedEntry('chrX', 3858266, 3858530)
        print(entry1 in bf)
        entry2 = bed.BedEntry('chrX', 10047550, 10067694)
        for x in bf.getOverlap(entry2):
            print(x)
        entry3 = bed.BedEntry('chr9', 102699903, 102700167)
        for x in bf.getClosest(entry3):
            print(x)
            for y in x:
                print(y)

    def test_entry_creation(self):
        entry1_1 = bed.BedEntry('X', 8000, 9000)
        entry1_1.signalValue = 10
        entry1_2 = bed.BedEntry('X', 80, 900)
        entry1_2.signalValue = 3
        bed1 = bed.BedFile([entry1_1, entry1_2])

        entry2_1 = bed.BedEntry('X', 8500, 9000)
        entry2_2 = bed.BedEntry('X', 80, 900)
        entry2_1.signalValue = 10
        entry2_2.signalValue = 3
        bed2 = bed.BedFile([entry2_1, entry2_2])

        entry3_1 = bed.BedEntry('X', 7500, 9000)
        entry3_2 = bed.BedEntry('X', 80, 900)
        entry3_1.signalValue = 10
        entry3_2.signalValue = 3
        bed3 = bed.BedFile([entry3_1, entry3_2])

        entry4_1 = bed.BedEntry('X', 5000, 8999)
        entry4_2 = bed.BedEntry('X', 80, 900)

        entry4_1.signalValue = 10
        entry4_2.signalValue = 3
        bed4 = bed.BedFile([entry4_1, entry4_2])

        unions = union([bed1, bed2, bed3, bed4], 2)
        for fragment in unions[0]:
            print(fragment)

    def test_single_chrom(self):
        med1 = bed.BedFile("test/data/med/med1_peaks.broadPeak", "Peaks").getChrom("chr17")
        med2 = bed.BedFile("test/data/med/med2_peaks.broadPeak", "Peaks").getChrom("chr17")
        med3 = bed.BedFile("test/data/med/med3_peaks.broadPeak", "Peaks").getChrom("chr17")

        bedf = [med1, med2, med3]
        minentries = 2

        # First create intersection and rank the entries in each replicate and return the rankproduct values
        ranks = rankreps(bedf, minentries, rankmethod='signalValue')

        # Calculate rank product for each entry that contributes to a union entry
        # Calculate the pvalues of the rank product values
        print('Calculating rank product probabilities...')
        rpb_up = rankprodbounds(ranks[1], len(ranks[1]), len(bedf), 'geometric')
        print('Calculating binomial threshold...')
        # Calculate rpb and binomial intersection point
        Pks = thresholdCalc(rpb_up, k=len(bedf) - (minentries - 1))
        if len(Pks[2]) != 0:
            binomAlpha = round(min(Pks[2]), 3)
        else:
            print('No binomial convergence, defaulting to 0.1')
            binomAlpha = 0.1

        # Perform multiple hypothesis testing correction upon the pvals
        fdr = multipletesting.fdrcorrection(rpb_up)

        # Determine whether to remove entries that are called significant
        print('Cleaning up output...')
        for i, v in enumerate(ranks[0][0]):
            p = rpb_up[i]

            if p != 0.0:
                ranks[0][0][i].addOption(name='TBD',
                                         score=min([abs(int(125 * math.log2(rpb_up[i]))), 1000]),
                                         strand='.',
                                         pValue=rpb_up[i],
                                         qValue=fdr[1][i])
            else:
                ranks[0][0][i].addOption(name='TBD',
                                         score=1000,
                                         strand='.',
                                         pValue=2.5e-20,
                                         qValue=2.5e-20)
        collapsed = bed.BedFile(ranks[0][0], 'IDR')

        connected = connect_entries(collapsed, bedf, 20, True)
        self.assertNotEqual(len(collapsed), len(connected))


    def test_from_bed(self):
        med1 = bed.BedFile("test/data/med/med1_peaks.broadPeak", "Peaks")
        med2 = bed.BedFile("test/data/med/med2_peaks.broadPeak", "Peaks")
        med3 = bed.BedFile("test/data/med/med3_peaks.broadPeak", "Peaks")

        bedf = [med1, med2, med3]

        performrankprod(bedf, minentries=2, rankmethod="pvalue", specifyMax=None,
                        duphandling='average', random_seed=0.5,
                        alpha=0.05,
                        filename="test_fragments_true",
                        default_min_peak=20,
                        print_pvals=True,
                        fragment=True)

        performrankprod(bedf, minentries=2, rankmethod="pvalue", specifyMax=None,
                        duphandling='average', random_seed=0.5,
                        alpha=0.05,
                        filename="test_fragments_false",
                        default_min_peak=20,
                        print_pvals=True,
                        fragment=False)

if __name__ == '__main__':
    unittest.main()


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ChIP-R',
    version='1.1.11',
    author='Rhys Newell, Mikael Boden, Alex Essebier',
    author_email='r.newell@uq.edu.au',
    packages=find_packages(),
    license='GPL-3.0',
    long_description_content_type="text/markdown",
    long_description=long_description,
    scripts=['bin/chipr'],
    install_requires = ['scipy', 'numpy'],
    entry_points={
        'console_scripts': [
        'chipr=chipr.__main__:main',
        'ChIP-R=chipr.__main__:main'
        ]
    },
    python_requires='>=3',
    description=('ChIP-R is a method for assessing the reproducibility of ' +
                 'replicated ChIP-seq type experiments. It incorporates the ' +
                 'rank product method, a novel thresholding methods, and the ' +
                 'use of peak fragmentation return the most reproducible peaks.'),
    keywords='ChIP-R',
    url='https://github.com/rhysnewell/ChIP-R',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
)

